"""
The module provides REST API handlers for some SCADA data requests.
"""
import warnings
import falcon

from ..base_handler import BaseHandler
from ..res_manager import ResourceManager
from ...simulation import SensorConfig, SensorFault


class ScadaDataManager(ResourceManager):
    """
    Class for managing SCADA data.
    """


class ScadaDataBaseHandler(BaseHandler):
    """
    Base class for all handlers concerning SCADA data.

    Parameters
    ----------
    scada_data_mgr : :class:`~epyt_flow.rest_api.scada_data.handlers.ScadaDataManager`
        SCADA data manager.
    """
    def __init__(self, scada_data_mgr: ScadaDataManager):
        self.scada_data_mgr = scada_data_mgr


class ScadaDataRemoveHandler(ScadaDataBaseHandler):
    """
    Class for handling a DELETE request for a given SCADA data instance.
    """
    def on_delete(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Deletes a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data instance.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            self.scada_data_mgr.remove(data_id)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataSensorConfigHandler(ScadaDataBaseHandler):
    """
    Class for handling GET and POST requests for the sensor configuration
    of a given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets the sensor configuration of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sensor_config = self.scada_data_mgr.get(data_id).sensor_config
            self.send_json_response(resp, my_sensor_config)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, data_id: str) -> None:
        """
        Sets the sensor configuration of a given SCADA data instance.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_config = self.load_json_data_from_request(req)
            if not isinstance(sensor_config, SensorConfig):
                self.send_json_parsing_error(resp)
                return

            self.scada_data_mgr.get(data_id).sensor_config = sensor_config
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataSensorFaultsHandler(ScadaDataBaseHandler):
    """
    Class for handling GET and POST requests concerning sensor faults in a
    given SCADA data instance.
    """
    def on_get(self, _, resp: falcon.Response, data_id: str) -> None:
        """
        Gets all sensor faults of a given SCADA data instance.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_faults = self.scada_data_mgr.get(data_id).sensor_faults
            self.send_json_response(resp, sensor_faults)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, data_id: str) -> None:
        """
        Sets (i.e. overrides) the sensor faults in a given SCADA data instance.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_faults = self.load_json_data_from_request(req)
            if not isinstance(sensor_faults, list) or \
                    any(not isinstance(e, SensorFault) for e in sensor_faults):
                self.send_json_parsing_error(resp)
                return

            self.scada_data_mgr.get(data_id).sensor_faults = sensor_faults
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScadaDataConvertUnitsHandler(ScadaDataBaseHandler):
    """
    Class for handling POST requests concerning unit conversion of a
    given SCADA data instance.
    """
    def on_post(self, req: falcon.Request, resp: falcon.Response, data_id: str) -> None:
        """
        Converts the units of a given SCADA data instance and returns a new SCADA data instance.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data_id : `str`
            UUID of the SCADA data.
        """
        try:
            if self.scada_data_mgr.validate_uuid(data_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            new_units = self.load_json_data_from_request(req)
            if not isinstance(new_units, dict):
                self.send_json_parsing_error(resp)
                return

            my_scada_data = self.scada_data_mgr.get(data_id)
            try:
                scada_data_new = my_scada_data.convert_units(**new_units)
            except Exception as ex:
                self.send_error(resp, str(ex))
                return

            new_scada_data_id = self.scada_data_mgr.create_new_item(scada_data_new)
            self.send_json_response(resp, {"data_id": new_scada_data_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
Module provides classes for implementing leakages.
"""
from copy import deepcopy
import math
import numpy as np
from epanet_plus import EPyT, EpanetConstants

from .system_event import SystemEvent
from ...serialization import serializable, JsonSerializable, \
    LEAKAGE_ID, ABRUPT_LEAKAGE_ID, INCIPIENT_LEAKAGE_ID


@serializable(LEAKAGE_ID, ".epytflow_leakage")
class Leakage(SystemEvent, JsonSerializable):
    """
    Base class for a leakage.

    Parameters
    ----------
    link_id : `str`
        ID of the link at which the leak is placed.
        Note that if the leak is placed at a node, then 'link_id' must be None and the
        ID of the node must be set in 'node_id'
    diameter : `float`, optional
        Diameter of this leak in either *foot* or *meter* (depending on the used flow units).

        Alternatively, 'area' can be used for specifying the size of this leakage --
        in this case, 'diameter' must be set to 'None'.

        The default is None.
    area : `float`, optional
        Area of this leak in either *foot^2* or *meter^2* (depending on the used flow units).

        Alternatively, 'diameter' can be used for specifying the size of this leakage --
        in this case, 'area' must be set to 'None'.

        The default is None.
    profile : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        Pattern of this leak.
    node_id : `str`, optional
        ID of the node at which the leak is placed.
        This parameter must only be set if the leak is placed at a node instead of a link.
        In this case, 'link_id' must be None.

        The default is None.
    """
    def __init__(self, link_id: str, profile: np.ndarray, diameter: float = None,
                 area: float = None, node_id: str = None, **kwds):
        if link_id is not None and node_id is not None:
            raise ValueError("Leak can not be placed at a link and node at the same time")
        if link_id is None and node_id is None:
            raise ValueError("Leak must be placed at either a link or a node -- " +
                             "expecting either 'link_id' or 'node_id' but both are None")
        if link_id is not None:
            if not isinstance(link_id, str):
                raise TypeError("'link_id' must be an instance of 'str' " +
                                f"but not of '{type(link_id)}'")
        if area is None and diameter is None:
            raise ValueError("Either 'diameter' or 'area' must be given")
        if area is not None and diameter is not None:
            raise ValueError("Either 'diameter' or 'area' must be given, " +
                             "but not both at the same time")
        if diameter is not None:
            if not isinstance(diameter, float):
                raise TypeError("'diameter must be an instance of 'float' but " +
                                f"not of '{type(diameter)}'")
            if diameter <= 0:
                raise ValueError("'diameter' must be greater than zero")
        if area is not None:
            if not isinstance(area, float):
                raise TypeError("'area must be an instance of 'float' but " +
                                f"not of '{type(area)}'")
            if area <= 0:
                raise ValueError("'area' must be greater than zero")
        if profile is not None:
            if not isinstance(profile, np.ndarray):
                raise TypeError("'profile' must be an instance of 'numpy.ndarray' but " +
                                f"not of '{type(profile)}'")
            if len(profile.shape) > 1:
                raise ValueError("'profile' must be a one-dimensional array " +
                                 f"but not of shape '{profile.shape}'")
        if node_id is not None:
            if not isinstance(node_id, str):
                raise TypeError("'node_id' must be an instance of 'str' " +
                                f"but not of '{type(node_id)}'")

        self.__link_id = link_id
        self.__node_id = node_id
        self.__diameter = diameter
        self.__area = area
        self.__profile = profile

        self.__leaky_node_idx = None
        self.__leak_emitter_coef = None
        self.__time_pattern_idx = 0

        super().__init__(**kwds)

    @property
    def link_id(self) -> str:
        """
        Gets the ID of the link at which the leak is placed.

        Returns
        -------
        `str`
            ID of the link at which the leak is placed.
        """
        return self.__link_id

    @property
    def node_id(self) -> str:
        """
        Gets the ID of the node at which the leak is placed.

        Returns
        -------
        `str`
            ID of the node at which the leak is placed.
        """
        return self.__node_id

    @property
    def diameter(self) -> float:
        """
        Gets the diameter of the leak in either *foot* or *meter*
        (depending on the sued flow units).

        Returns
        -------
        `float`
            Diameter (*foot* or *meter*) of the leak.
        """
        return self.__diameter

    @property
    def area(self) -> float:
        """
        Gets the area of the leak in either *foot^2* or *meter^2*
        (depending on the sued flow units).

        Returns
        -------
        `float`
            Area of the leak.
        """
        return self.__area if self.__area is not None else self.compute_leak_area(self.__diameter)

    @property
    def profile(self) -> np.ndarray:
        """
        Gets the pattern of the leak.

        Returns
        -------
        `numpy.ndarray`
            Pattern of the leak.
        """
        return deepcopy(self.__profile)

    @profile.setter
    def profile(self, pattern: np.ndarray):
        if not isinstance(pattern, np.ndarray):
            raise TypeError("'profile' must be an instance of 'numpy.ndarray' but " +
                            f"not of '{type(pattern)}'")
        if len(pattern.shape) > 1:
            raise ValueError("'profile' must be a one-dimensional array " +
                             f"but not of shape '{pattern.shape}'")

        self.__profile = pattern

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"link_id": self.__link_id, "diameter": self.__diameter,
                                           "area": self.__area, "profile": self.__profile,
                                           "node_id": self.__node_id
                                           if self.__link_id is None else None}

    def __eq__(self, other) -> bool:
        if not isinstance(other, Leakage):
            raise TypeError(f"Can not compare 'Leakage' instance with '{type(other)}' instance")

        return super().__eq__(other) and self.__link_id == other.link_id \
            and self.__diameter == other.diameter and np.all(self.__profile == other.profile) \
            and self.__node_id == other.node_id and self.area == other.area

    def __str__(self) -> str:
        return f"{super().__str__()} link_id: {self.__link_id} diameter: {self.__diameter} " +\
            f"area: {self.__area} profile: {self.__profile} node_id: {self.__node_id}"

    def compute_leak_area(self, diameter: float) -> float:
        """
        Computes the leak area given the diameter.

        leak_area = pi * (diameter * .5)^2

        Parameters
        ----------
        diameter : `float`
            Diameter (*foot* or *meter*) of the leak.

        Returns
        -------
        `float`
            Leak area in *foot^2* or *meter^2*.
        """
        return np.pi * (diameter / 2) ** 2

    def compute_leak_emitter_coefficient(self, area: float, discharge_coef: float = .75) -> float:
        """
        Computes the leak emitter coefficient.

        emitter_coef = discharge_coef * area * sqrt(2*g)
        where g is the gravitational constant, and discharge_coef = .75

        leak_demand = emitter_coef * pressure^alpha       where alpha = .5

        Parameters
        ----------
        area : `float`
            Leak area (foot^2 or meter^2) as computed in
            :func:`~epyt_flow.simulation.events.leakages.Leakage.compute_leak_area`.
        discharge_coef : `float`, optional
            Discharge coefficient.

            The default is set to 0.75

        Returns
        -------
        `float`
            Leak emitter coefficient.
        """
        flow_unit = self._epanet_api.getflowunits()
        if flow_unit == EpanetConstants.EN_CMH:
            g = 127137600   # m/h^2
        elif flow_unit == EpanetConstants.EN_CFS:
            g = 32.17405    # feet/s^2
        else:
            raise ValueError("Leakages are only implemented for the following flow units:\n" +
                             " EN_CMH (cubic meter/hr)\n EN_CFS (foot/sec)")

        return discharge_coef * area * np.sqrt(2. * g)

    def _get_new_link_id(self) -> str:
        return f"leak_pipe_{self.__link_id}"

    def _get_new_node_id(self) -> str:
        return f"leak_node_{self.__link_id}"

    def init(self, epanet_api: EPyT) -> None:
        super().init(epanet_api)

        # Split pipe if leak is placed at a link/pipe
        if self.__link_id is not None:
            if self.__link_id not in self._epanet_api.get_all_links_id():
                raise ValueError(f"Unknown link/pipe '{self.__link_id}'")

            new_link_id = self._get_new_link_id()
            new_node_id = self._get_new_node_id()

            all_nodes_id = self._epanet_api.get_all_nodes_id()
            if new_node_id in all_nodes_id:
                raise ValueError(f"There is already a leak at pipe {self.link_id}")

            self._epanet_api.split_pipe(self.link_id, new_link_id, new_node_id)
            self.__leaky_node_idx = self._epanet_api.get_node_idx(new_node_id)
        else:
            if self.__node_id not in self._epanet_api.get_all_nodes_id():
                raise ValueError(f"Unknown node '{self.__node_id}'")

            self.__leaky_node_idx = self._epanet_api.get_node_idx(self.__node_id)

        self._epanet_api.setnodevalue(self.__leaky_node_idx, EpanetConstants.EN_EMITTER, 0.)

        # Compute leak emitter coefficient
        self.__leak_emitter_coef = self.compute_leak_emitter_coefficient(
            self.compute_leak_area(self.area))

    def cleanup(self) -> None:
        if self.__link_id is not None:
            pipe_idx = self._epanet_api.get_link_idx(self.__link_id)
            link_diameter = self._epanet_api.get_link_diameter(pipe_idx)
            link_length = self._epanet_api.get_link_length(pipe_idx)
            link_roughness_coeff = self._epanet_api.get_link_roughness(pipe_idx)
            link_minor_loss_coeff = self._epanet_api.get_link_minorloss(pipe_idx)
            link_initial_status = self._epanet_api.get_link_init_status(pipe_idx)
            link_initial_setting = self._epanet_api.get_link_init_setting(pipe_idx)
            link_bulk_reaction_coeff = self._epanet_api.get_link_bulk_raction_coeff(pipe_idx)
            link_wall_reaction_coeff = self._epanet_api.get_link_wall_raction_coeff(pipe_idx)

            node_a_idx = int(self._epanet_api.getlinknodes(pipe_idx)[0])
            node_b_idx = int(self._epanet_api.getlinknodes(self._epanet_api.get_link_idx(self._get_new_link_id()))[1])

            self._epanet_api.deletelink(self._epanet_api.get_link_idx(self._get_new_link_id()),
                                        EpanetConstants.EN_UNCONDITIONAL)
            self._epanet_api.deletelink(self._epanet_api.get_link_idx(self.__link_id),
                                        EpanetConstants.EN_UNCONDITIONAL)
            self._epanet_api.deletenode(self._epanet_api.get_node_idx(self._get_new_node_id()),
                                        EpanetConstants.EN_UNCONDITIONAL)

            self._epanet_api.addlink(self.__link_id, EpanetConstants.EN_PIPE,
                                     self._epanet_api.get_node_id(node_a_idx),
                                     self._epanet_api.get_node_id(node_b_idx))
            link_idx = self._epanet_api.get_link_idx(self.__link_id)
            self._epanet_api.setlinknodes(link_idx, node_a_idx, node_b_idx)
            self._epanet_api.setlinktype(link_idx, EpanetConstants.EN_PIPE,
                                         EpanetConstants.EN_UNCONDITIONAL)
            self._epanet_api.setpipedata(link_idx, link_length, link_diameter, link_roughness_coeff,
                                         link_minor_loss_coeff)
            self._epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_INITSTATUS,
                                          link_initial_status)
            self._epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_INITSETTING,
                                          link_initial_setting)
            self._epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_KBULK,
                                          link_bulk_reaction_coeff)
            self._epanet_api.setlinkvalue(link_idx, EpanetConstants.EN_KWALL,
                                          link_wall_reaction_coeff)

    def reset(self) -> None:
        self.__time_pattern_idx = 0

    def exit(self, cur_time) -> None:
        self._epanet_api.setnodevalue(self.__leaky_node_idx, EpanetConstants.EN_EMITTER, 0.)

    def apply(self, cur_time: int) -> None:
        self._epanet_api.setnodevalue(self.__leaky_node_idx, EpanetConstants.EN_EMITTER,
                                      self.__leak_emitter_coef *
                                      self.__profile[self.__time_pattern_idx])
        self.__time_pattern_idx += 1


@serializable(ABRUPT_LEAKAGE_ID, ".epytflow_leakage_abrupt")
class AbruptLeakage(Leakage):
    """
    Class implementing an abrupt leakage event.

    Parameters
    ----------
    link_id : `str`
        ID of the link at which the leak is placed.
    diameter : `float`, optional
        Diameter of the leak.

        Alternatively, 'area' can be used to specify the size of this leak --
        in this case, 'diameter' must be set to None.

        The default is None.
    area : `float`, optional
        Area of the leakd.

        Alternatively, 'diameter' can be used to specify the size of this leak --
        in this case, 'area' must be set to None.

        The default is None.
    """
    def __init__(self, link_id: str, diameter: float = None, area: float = None, **kwds):
        if "profile" not in kwds:
            super().__init__(link_id=link_id, diameter=diameter, area=area, profile=None, **kwds)
        else:
            super().__init__(link_id=link_id, diameter=diameter, area=area, **kwds)

    def init(self, epanet_api: EPyT) -> None:
        super().init(epanet_api)

        # Set pattern
        total_sim_duration = self._epanet_api.get_simulation_duration()
        time_step = self._epanet_api.get_hydraulic_time_step()

        if self.end_time is not None:
            n_leaky_time_points = math.ceil((self.end_time - self.start_time) / time_step)
        else:
            n_leaky_time_points = math.ceil((total_sim_duration - self.start_time) / time_step)

        self.profile = np.ones(n_leaky_time_points)


@serializable(INCIPIENT_LEAKAGE_ID, ".epytflow_leakage_incipient")
class IncipientLeakage(Leakage):
    """
    Class implementing an incipient leakage event.

    Parameters
    ----------
    link_id : `str`
        ID of the link at which the leak is placed.
    diameter : `float`, optional
        Maximum diameter of the leak -- i.e. small leak diameter in the beginning,
        growing over time until peak time is reached.

        Alternatively, 'area' can be used to specify the size of this leak --
        in this case, 'diameter' must be set to None.

        The default is None.
    area : `float`, optional
        Maximum area of the leak -- i.e. small leak area in the beginning,
        growing over time until peak time is reached.

        Alternatively, 'diameter' can be used to specify the size of this leak --
        in this case, 'area' must be set to None.

        The default is None.
    peak_time : `int`
        Time (seconds since the simulation start) when this leak reaches
        its larges size (leak diameter).
    """
    def __init__(self, link_id: str, peak_time: int, diameter: float = None,
                 area: float = None, **kwds):
        if peak_time < kwds["start_time"] or (kwds["end_time"] is not None and
                                              peak_time > kwds["end_time"]):
            raise ValueError("'peak_time' must be greater than 'start_time' and " +
                             "smaller than 'end_time'")

        self.__peak_time = peak_time

        if "profile" not in kwds:
            super().__init__(link_id=link_id, diameter=diameter, area=area, profile=None, **kwds)
        else:
            super().__init__(link_id=link_id, diameter=diameter, area=area, **kwds)

    @property
    def peak_time(self) -> int:
        """
        Gets the peak time (seconds since the simulation start) of the leak.

        Returns
        -------
        `int`
            Peak time of the leak.
        """
        return self.__peak_time

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"peak_time": self.peak_time}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.peak_time == other.peak_time

    def __str__(self) -> str:
        return f"{super().__str__()} peak_time: {self.peak_time}"

    def init(self, epanet_api: EPyT) -> None:
        super().init(epanet_api)

        # Set pattern
        total_sim_duration = self._epanet_api.get_simulation_duration()
        time_step = self._epanet_api.get_hydraulic_time_step()

        if self.end_time is not None:
            n_leaky_time_points = math.ceil((self.end_time - self.start_time) / time_step)
        else:
            n_leaky_time_points = math.ceil((total_sim_duration - self.start_time) / time_step)

        profile = np.ones(n_leaky_time_points)

        coeff = int((self.peak_time - self.start_time) / time_step)
        for t in range(coeff):
            profile[t] = (1. / coeff) + ((1. / coeff) * t)    # Linear interpolation!

        self.profile = profile



"""
This module provides REST API handlers concerning the simulation of scenarios.
"""
import warnings
import falcon

from .handlers import ScenarioBaseHandler
from ..scada_data.handlers import ScadaDataManager


class ScenarioSimulationHandler(ScenarioBaseHandler):
    """
    Class for handling GET requests for simulating a given scenario.

    Parameters
    ----------
    scada_data_mgr : :class:`~epyt_flow.rest_api.scada_data.handlers.ScadaDataManager`
        SCADA data manager.
    """
    def __init__(self, scada_data_mgr: ScadaDataManager, **kwds):
        self.scada_data_mgr = scada_data_mgr

        super().__init__(**kwds)

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Runs the simulation of a given scenario.

        Note that in contrat to the GET request
        (:func:`~epyt_flow.rest_api.scenario.simulation_handlers.ScenarioSimulationHandler.on_get`),
        the POST request allows to specify additional arguments passed to
        :func:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator.run_simulation`.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            params = self.load_json_data_from_request(req)

            my_scenario = self.scenario_mgr.get(scenario_id)
            res = my_scenario.run_simulation(**params)

            data_id = self.scada_data_mgr.create_new_item(res)
            self.send_json_response(resp, {"data_id": data_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.data = str(ex)
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Runs the simulation of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_scenario = self.scenario_mgr.get(scenario_id)
            res = my_scenario.run_simulation()

            data_id = self.scada_data_mgr.create_new_item(res)
            self.send_json_response(resp, {"data_id": data_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.data = str(ex)
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioBasicQualitySimulationHandler(ScenarioBaseHandler):
    """
    Class for handling POST requests for runing a basic quality simulation of a given scenario.

    Parameters
    ----------
    scada_data_mgr : :class:`~epyt_flow.rest_api.scada_data.handlers.ScadaDataManager`
        SCADA data manager.
    """
    def __init__(self, scada_data_mgr: ScadaDataManager, **kwds):
        self.scada_data_mgr = scada_data_mgr

        super().__init__(**kwds)

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Runs the basic quality simulation of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            params = self.load_json_data_from_request(req)

            my_scenario = self.scenario_mgr.get(scenario_id)
            res = my_scenario.run_basic_quality_simulation(**params)

            data_id = self.scada_data_mgr.create_new_item(res)
            self.send_json_response(resp, {"data_id": data_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.data = str(ex)
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioAdvancedQualitySimulationHandler(ScenarioBaseHandler):
    """
    Class for handling POST requests for runing an advanced quality simulation of a given scenario.

    Parameters
    ----------
    scada_data_mgr : :class:`~epyt_flow.rest_api.scada_data.handlers.ScadaDataManager`
        SCADA data manager.
    """
    def __init__(self, scada_data_mgr: ScadaDataManager, **kwds):
        self.scada_data_mgr = scada_data_mgr

        super().__init__(**kwds)

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Runs the advanced quality simulation of a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            params = self.load_json_data_from_request(req)

            my_scenario = self.scenario_mgr.get(scenario_id)
            res = my_scenario.run_advanced_quality_simulation(**params)

            data_id = self.scada_data_mgr.create_new_item(res)
            self.send_json_response(resp, {"data_id": data_id})
        except Exception as ex:
            warnings.warn(str(ex))
            resp.data = str(ex)
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
This module provides classes for implementing different types of sensor reading attacks.
"""
from copy import deepcopy
import numpy as np

from .sensor_reading_event import SensorReadingEvent
from ...serialization import serializable, JsonSerializable, SENSOR_ATTACK_OVERRIDE_ID, \
    SENSOR_ATTACK_REPLAY_ID


class SensorReadingAttack(SensorReadingEvent):
    """
    Base class of a sensor reading attack.
    """


@serializable(SENSOR_ATTACK_OVERRIDE_ID, ".epytflow_sensorattack_override")
class SensorOverrideAttack(SensorReadingAttack, JsonSerializable):
    """
    Class implementing a sensor override attack -- i.e. sensor reading values are overwritten
    by pre-defined values.

    If the override attack is running out of pre-defined sensor reading values,
    it repeats the given values from the beginning onwards.

    Parameters
    ----------
    new_sensor_values : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
        New sensor reading values -- i.e. these values replace the true sensor reading values.
    """
    def __init__(self, new_sensor_values: np.ndarray, **kwds):
        if not isinstance(new_sensor_values, np.ndarray):
            raise TypeError("'new_sensor_values' must be an instance of 'numpy.ndarray' " +
                            f"but not of '{type(new_sensor_values)}'")
        if len(new_sensor_values.shape) != 1:
            raise ValueError("'new_sensor_values' must be a 1-dimensional array")
        if len(new_sensor_values) == 0:
            raise ValueError("'new_sensor_values' can not be empty")

        self.__new_sensor_values = new_sensor_values
        self.__cur_replay_idx = 0

        super().__init__(**kwds)

    @property
    def new_sensor_values(self) -> np.ndarray:
        """
        Returns the new sensor reading values -- i.e. these values replace the
        true sensor reading values.

        Returns
        -------
        `np.ndarray`
            New sensor readings.
        """
        return deepcopy(self.__new_sensor_values)

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"new_sensor_values": self.__new_sensor_values}

    def __eq__(self, other) -> bool:
        if not isinstance(other, SensorOverrideAttack):
            raise TypeError("Can not compare 'SensorOverrideAttack' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and np.all(self.__new_sensor_values == other.new_sensor_values)

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} " +\
            f"new_sensor_values: {self.__new_sensor_values}"

    def apply(self, sensor_readings: np.ndarray,
              sensor_readings_time: np.ndarray) -> np.ndarray:
        for i in range(sensor_readings.shape[0]):
            t = sensor_readings_time[i]

            if self.start_time <= t <= self.end_time:
                sensor_readings[i] = self.__new_sensor_values[self.__cur_replay_idx]
                self.__cur_replay_idx = (self.__cur_replay_idx + 1) % len(self.__new_sensor_values)

        return sensor_readings


@serializable(SENSOR_ATTACK_REPLAY_ID, ".epytflow_sensorattack_replay")
class SensorReplayAttack(SensorReadingAttack, JsonSerializable):
    """
    Class implementing a sensor replay attack -- i.e. sensor readings are replaced by
    historical recordings.

    If the provided time window of historical recordings is smaller than the time window of the
    attack, it repeats the historical values from the beginning onwards.

    Parameters
    ----------
    replay_data_time_window_start : `int`
        Start (seconds since simulation start) of the time window that is used in the replay
        of sensor readings.
    replay_data_time_window_end : `int`
        End (seconds since simulation start) of the time window that is used in the replay
        of sensor readings.
    """
    def __init__(self, replay_data_time_window_start: int, replay_data_time_window_end: int,
                 **kwds):
        if not isinstance(replay_data_time_window_start, int):
            raise TypeError("'replay_data_time_window_start' must be an instance of 'int' " +
                            f"but not of {type(replay_data_time_window_start)}")
        if not isinstance(replay_data_time_window_end, int):
            raise TypeError("'replay_data_time_window_end' must be an instance of 'int' " +
                            f"but not of {type(replay_data_time_window_end)}")
        if replay_data_time_window_start > replay_data_time_window_end or \
                replay_data_time_window_start < 0:
            raise ValueError("Invalid values for 'replay_data_time_window_start' and/or " +
                             "'replay_data_time_window_end' detected.")

        self.__new_sensor_values = np.zeros(replay_data_time_window_end -
                                            replay_data_time_window_start)
        self.__sensor_data_time_window_start = replay_data_time_window_start
        self.__sensor_data_time_window_end = replay_data_time_window_end
        self.__cur_hist_idx = 0
        self.__cur_replay_idx = 0

        super().__init__(**kwds)

        if self.__sensor_data_time_window_start > self.start_time:
            raise ValueError("'replay_data_time_window_start' must be less than 'start_time'")

    @property
    def sensor_data_time_window_start(self) -> int:
        """
        Gets the start time (seconds since simulation start) of the time window
        that is used in the replay of sensor readings.

        Returns
        -------
        `int`
            Start time.
        """
        return self.__sensor_data_time_window_start

    @property
    def sensor_data_time_window_end(self) -> int:
        """
        Gets the end time (seconds since simulation start) of the time window
        that is used in the replay of sensor readings.

        Returns
        -------
        `int`
            End time.
        """
        return self.__sensor_data_time_window_end

    @property
    def new_sensor_values(self) -> np.ndarray:
        """
        Returns the new sensor reading values -- i.e. these values replace the
        true sensor reading values.

        Returns
        -------
        `np.ndarray`
            New sensor readings.
        """
        return deepcopy(self.__new_sensor_values)

    def get_attributes(self) -> dict:
        my_attributes = {"new_sensor_values": self.__new_sensor_values,
                         "replay_data_time_window_start": self.__sensor_data_time_window_start,
                         "replay_data_time_window_end": self.__sensor_data_time_window_end}

        return super().get_attributes() | my_attributes

    def __eq__(self, other) -> bool:
        if not isinstance(other, SensorReplayAttack):
            raise TypeError("Can not compare 'SensorReplayAttack' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and np.all(self.__new_sensor_values == other.new_sensor_values)

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} " +\
            f"new_sensor_values: {self.__new_sensor_values}"

    def apply(self, sensor_readings: np.ndarray,
              sensor_readings_time: np.ndarray) -> np.ndarray:
        for i in range(sensor_readings.shape[0]):
            t = sensor_readings_time[i]

            if self.__sensor_data_time_window_start <= t <= self.__sensor_data_time_window_end:
                self.__new_sensor_values[self.__cur_hist_idx] = sensor_readings[i]
                self.__cur_hist_idx += 1

            if self.start_time <= t <= self.end_time:
                sensor_readings[i] = self.__new_sensor_values[self.__cur_replay_idx]
                self.__cur_replay_idx = (self.__cur_replay_idx + 1) % len(self.__new_sensor_values)

        return sensor_readings



"""
The BATtle of the Attack Detection ALgorithms (*BATADAL*) by Riccardo Taormina, Stefano Galelli,
Nils Ole Tippenhauer, Avi Ostfeld, Elad Salomons, Demetrios Eliades is a competition on planning
and management of water networks undertaken within the Water Distribution Systems Analysis
Symposium. The goal of the battle was to compare the performance of algorithms for the detection
of cyber-physical attacks, whose frequency has increased in the last few years along with the
adoption of smart water technologies. The design challenge was set for the C-Town network,
a real-world, medium-sized water distribution system operated through programmable logic
controllers and a supervisory control and data acquisition (SCADA) system. Participants were
provided with data sets containing (simulated) SCADA observations, and challenged to design
an attack detection algorithm. The effectiveness of all submitted algorithms was evaluated in
terms of time-to-detection and classification accuracy. Seven teams participated in the battle
and proposed a variety of successful approaches leveraging data analysis, model-based detection
mechanisms, and rule checking. Results were presented at the Water Distribution Systems Analysis
Symposium (World Environmental and Water Resources Congress) in Sacramento, California on
May 21-25, 2017.
The `paper <https://doi.org/10.1061/(ASCE)WR.1943-5452.0000969>`_ summarizes the BATADAL
problem, proposed algorithms, results, and future research directions.

See https://www.batadal.net/ for details.

This module provides functions for loading the original BATADAL data set
:func:`~epyt_flow.data.benchmarks.batadal.load_data`, as well as functions for loading the
scenarios :func:`~epyt_flow.data.benchmarks.batadal.load_scenario` and pre-generated
SCADA data :func:`~epyt_flow.data.benchmarks.batadal.load_scada_data`.
"""
import os
from typing import Any
from datetime import datetime
import pandas as pd
import numpy as np

from .batadal_data import TRAINING_DATA_2_ATTACKS_TIME, TRAINING_DATA_2_START_TIME, \
    TEST_DATA_ATTACKS_TIME, TEST_DATA_START_TIME
from ...utils import get_temp_folder, unpack_zip_archive, to_seconds, download_if_necessary
from ...simulation import ScenarioConfig


def __parse_attacks_time(start_time: str, attacks_time):
    events = []
    for event in attacks_time.splitlines():
        # Parse entry
        items = [i.strip() for i in event.split(",")]

        event_start_time = int((datetime.strptime(items[0], "%d/%m/%Y %H:%M") - start_time)
                               .total_seconds())
        event_end_time = int((datetime.strptime(items[1], "%d/%m/%Y %H:%M") - start_time)
                             .total_seconds())

        events.append((event_start_time, event_end_time))

    return events


def load_data(download_dir: str = None, return_X_y: bool = False,
              return_ground_truth: bool = False, return_features_desc: bool = False,
              verbose: bool = True) -> dict:
    """
    Loads the original BATADAL competition data.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    return_X_y : `bool`, optional
        If True, the data together with the labels is returned as pairs of Numpy arrays.
        Otherwise, the data is returned as Pandas data frames.

        The default is False.
    return_ground_truth : `bool`
        If True and if `return_X_y` is True, the ground truth labels are included in the
        returned dictionary -- note that the labels provided in the benchmark constitute
        a partial labeling only.

        The default is False.
    return_features_desc : `bool`
        If True and if `return_X_y` is True, feature names (i.e. descriptions) are included
        in the returned dictionary.

        The default is False.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    `dict`
        Dictionary of the loaded benchmark data. The dictionary contains the two training
        data sets ("train_1" and "train_2"), as well as the test data set ("test").
        If `return_X_y` is False, each dictionary entry is a `Pandas dataframe <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_.
        Otherwise, it is a tuple of sensor readings and labels (except for the test set) --
        if `return_ground_truth` is True or `return_features_desc` is True, the corresponding
        data is appended to the tuple.
    """
    download_dir = download_dir if download_dir is not None else get_temp_folder()

    # Download data
    training_data_1_url = "https://www.batadal.net/data/BATADAL_dataset03.csv"
    training_data_2_url = "https://www.batadal.net/data/BATADAL_dataset04.csv"
    test_data_url = "https://www.batadal.net/data/BATADAL_test_dataset.zip"

    training_data_1_path = os.path.join(download_dir, "BATADAL_dataset03.csv")
    training_data_2_path = os.path.join(download_dir, "BATADAL_dataset04.csv")

    download_if_necessary(training_data_1_path, training_data_1_url, verbose)
    download_if_necessary(training_data_2_path, training_data_2_url, verbose)

    download_if_necessary(os.path.join(download_dir, "BATADAL_test_dataset.zip"),
                          test_data_url, verbose)
    unpack_zip_archive(os.path.join(download_dir, "BATADAL_test_dataset.zip"), download_dir)

    # Load and return data
    df_train_1 = pd.read_csv(training_data_1_path)
    df_train_2 = pd.read_csv(training_data_2_path)
    df_test = pd.read_csv(os.path.join(download_dir, "BATADAL_test_dataset.csv"))

    if return_X_y is True:
        # Convert data to numpy
        y_train_1 = df_train_1["ATT_FLAG"].to_numpy().astype(np.int8)
        del df_train_1["ATT_FLAG"]
        del df_train_1["DATETIME"]
        X_train_1 = df_train_1.to_numpy()

        y_train_2 = df_train_2[" ATT_FLAG"].to_numpy()
        idx = np.argwhere(y_train_2 == -999)
        y_train_2[idx] = 0
        y_train_2 = y_train_2.astype(np.int8)
        del df_train_2[" ATT_FLAG"]
        del df_train_2["DATETIME"]
        X_train_2 = df_train_2.to_numpy()

        del df_test["DATETIME"]
        X_test = df_test.to_numpy()

        # Create ground truth labels
        hydraulic_time_step = to_seconds(minutes=15)
        training_data_2_events_time = __parse_attacks_time(TRAINING_DATA_2_START_TIME,
                                                           TRAINING_DATA_2_ATTACKS_TIME)
        test_data_events_time = __parse_attacks_time(TEST_DATA_START_TIME, TEST_DATA_ATTACKS_TIME)

        y_train_2_truth = np.zeros(X_train_2.shape[0])
        for event_start, event_end in training_data_2_events_time:
            t0 = int(event_start / hydraulic_time_step)
            t1 = int(event_end / hydraulic_time_step)
            y_train_2_truth[t0:t1] = 1

        y_test_truth = np.zeros(X_test.shape[0])
        for event_start, event_end in test_data_events_time:
            t0 = int(event_start / hydraulic_time_step)
            t1 = int(event_end / hydraulic_time_step)
            y_test_truth[t0:t1] = 1

        # Create features' descriptions
        features_desc = list(df_train_1.columns)
        desc_mapping = {"PU": "Pump", "V": "Valve", "T": "Tank", "L": "Level", "S": "State",
                        "P": "Pressure", "F": "Flow"}
        for i, f_desc in enumerate(features_desc):
            pump = False
            for k, value in desc_mapping.items():
                if k in f_desc:
                    if k == "P" and pump is True:
                        continue
                    f_desc = f_desc.replace(k, value)
                    if k == "PU":
                        pump = True
            features_desc[i] = f_desc

        # Create final results
        r = {"train_1": (X_train_1, y_train_1), "train_2": (X_train_2, y_train_2),
             "test": X_test}

        if return_ground_truth is True:
            r["train_1"] = (r["train_1"][0], r["train_1"][1], y_train_1)
            r["train_2"] = (r["train_2"][0], r["train_2"][1], y_train_2_truth)
            r["test"] = (r["test"][0], y_test_truth)

        if return_features_desc is True:
            r["features_desc"] = features_desc

        return r
    else:
        return {"train_1": df_train_1, "train_2": df_train_2, "test": df_test}


def load_scada_data(download_dir: str = None, return_X_y: bool = False,
                    return_ground_truth: bool = False, return_features_desc: bool = False,
                    verbose: bool = True) -> Any:
    """
    Loads the SCADA data of the simulated BATADAL benchmark scenario -- note that due to
    randomness and undocumented aspects of the original BATADAL data set, these differ from
    the original data set which can be loaded by calling
    :func:`~epyt_flow.data.benchmarks.batadal.load_data`.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    return_X_y : `bool`, optional
        If True, the data together with the labels is returned as pairs of Numpy arrays.
        Otherwisen the data is returned as Pandas data frames.

        The default is False.
    return_ground_truth : `bool`
        If True and if `return_X_y` is True, the ground truth labels are included in the
        returned dictionary -- note that the labels provided in the benchmark constitute
        a partial labeling only.

        The default is False.
    return_features_desc : `bool`
        If True and if `return_X_y` is True, feature names (i.e. descriptions) are included
        in the returned dictionary.

        The default is False.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.
    """
    raise NotImplementedError()


def load_scenario(download_dir: str = None, verbose: bool = True) -> ScenarioConfig:
    """
    Creates and returns the BATADAL scenario -- it can be either modified or directly passed
    to the simulator :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.

    .. note::

        Note that due to randomness and undocumented aspects of the original BATADAL benchmark,
        the scenario simulation results differ from the original data set which can be loaded by
        calling :func:`~epyt_flow.data.benchmarks.batadal.load_data`.

    Parameters
    ----------
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        The BATADAL scenario.
    """
    raise NotImplementedError()



"""
Module provides a class for specifying scenario configurations.
"""
from typing import Any
from copy import deepcopy
import os
import json
from pathlib import Path
import numpy as np

from ..uncertainty import AbsoluteGaussianUncertainty, RelativeGaussianUncertainty, \
    AbsoluteUniformUncertainty, RelativeUniformUncertainty, ModelUncertainty, \
    SensorNoise, Uncertainty
from .sensor_config import SensorConfig
from .scada import CustomControlModule, SimpleControlModule, ComplexControlModule
from .events import SystemEvent, SensorReadingEvent
from .events.sensor_faults import SensorFaultConstant, SensorFaultDrift, SensorFaultGaussian, \
    SensorFaultPercentage, SensorFaultStuckZero
from .events.leakages import AbruptLeakage, IncipientLeakage
from ..serialization import serializable, Serializable, SCENARIO_CONFIG_ID


@serializable(SCENARIO_CONFIG_ID, ".epytflow_scenario_config")
class ScenarioConfig(Serializable):
    """
    Configuration of a scenario.

    Parameters
    ----------
    scenario_config : :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`, optional
        Uses the given scenario configuration to create this instance --
        other attributes passed to this constructor override the attributes in 'scenario_config'.

        Note that if 'scenario_config' is None then 'f_inp_in' can not be None --
        i.e. either 'scenario_config' or 'f_inp_in' must be given.

        The default is None.
    f_inp_in : `str`, optional
        Path to the .inp file.

        Note that if 'f_inp_in' is None then 'scenario_config' can not be None --
        i.e. either 'scenario_config' or 'f_inp_in' must be given.

        The default is None.
    f_msx_in : `str`, optional
        Path to the .msx file -- optional, only necessary if EPANET-MSX is used.

        The default is None
    general_params : `dict`, optional
        General parameters such as the demand model, hydraulic time steps, etc.

        The default is None
    sensor_config : :class:`~epyt_flow.simulation.sensor_config.SensorConfig`, optional
        Specification of all sensors.

        The default is None
    memory_consumption_estimate : float, optional
        Estimated memory consumption of this scenario in MB -- i.e. the amount of memory that is
        needed on the hard disk as well as in RAM.

        The default is None.
    sensor_noise : :class:`~epyt_flow.uncertainty.sensor_noise.SensorNoise`, optional
        Speciation of sensor noise -- i.e. noise/uncertainty affecting the sensor readings.

        The default is None
    csutom_controls : list[:class:`~epyt_flow.simulation.scada.custom_control.CustomControlModule`], optional
        List of custom control modules that are active during the simulation.

        The default is an empty list.
    simple_controls : list[:class:`~epyt_flow.simulation.scada.simple_control.SimpleControlModule`], optional
        List of EPANET control rules that are active during the simulation.

        The default is an empty list.
    complex_controls : list[:class:`~epyt_flow.simulation.scada.complex_control.ComplexControlModule`], optional
        List of complex (i.e. IF-THEN-ELSE) EPANET control rules that are active during the simulation.

        The default is an empty list.
    model_uncertainty : :class:`~epyt_flow.uncertainty.model_uncertainty.ModelUncertainty`, optional
        Specification of model uncertainty.
    system_events : list[:class:`~epyt_flow.simulation.events.system_event.SystemEvent`], optional
        List of system events -- i.e. events that directly affect the simulation (e.g. leakages).

        The default is an empty list.
    sensor_reading_events : list[:class:`~epyt_flow.simulation.events.sensor_reading_event.SensorReadingEvent`], optional
        List of sensor reading events -- i.e. events that affect the readings of sensors.

        The default is an empty list.
    """

    def __init__(self, scenario_config: Any = None, f_inp_in: str = None, f_msx_in: str = None,
                 general_params: dict = None, sensor_config: SensorConfig = None,
                 memory_consumption_estimate: float = None,
                 custom_controls: list[CustomControlModule] = [],
                 simple_controls: list[SimpleControlModule] = [],
                 complex_controls: list[ComplexControlModule] = [],
                 sensor_noise: SensorNoise = None,
                 model_uncertainty: ModelUncertainty = None,
                 system_events: list[SystemEvent] = [],
                 sensor_reading_events: list[SensorReadingEvent] = [], **kwds):
        if f_inp_in is None and scenario_config is None:
            raise ValueError("Either 'f_inp_in' or 'scenario_config' must be given")
        if scenario_config is not None:
            if not isinstance(scenario_config, ScenarioConfig):
                raise TypeError("'scenario_config' must be an instance of " +
                                "'epyt_flow.simulation.ScenarioConfig' but not of " +
                                f"'{type(scenario_config)}'")
        if f_inp_in is not None:
            if not isinstance(f_inp_in, str):
                raise TypeError("'f_inp_in' must be an instance of 'str' " +
                                f"but no of '{type(f_inp_in)}'")
        if f_msx_in is not None:
            if not isinstance(f_msx_in, str):
                raise TypeError("'f_msx_in' must be an instance of 'str' " +
                                f"but no of '{type(f_msx_in)}'")
        if general_params is not None:
            if not isinstance(general_params, dict):
                raise TypeError("'general_params' must be an instance of 'dict' " +
                                f"but not of '{type(general_params)}'")
        if sensor_config is not None:
            if not isinstance(sensor_config, SensorConfig):
                raise TypeError("'sensor_config' must be an instance of " +
                                "'epyt_flow.simulation.SensorConfig' but not of " +
                                f"'{type(sensor_config)}'")
        if memory_consumption_estimate is not None:
            if not isinstance(memory_consumption_estimate, float) or \
                    memory_consumption_estimate <= 0:
                raise ValueError("'memory_consumption_estimate' must be a positive integer")
        if len(custom_controls) != 0:
            if any(not isinstance(c, CustomControlModule) for c in custom_controls):
                raise TypeError("Each item in 'custom_controls' must be an instance of " +
                                "'epyt_flow.simulation.scada.CustomControlModule'")
        if not isinstance(simple_controls, list):
            raise TypeError("'simple_controls' must be an instance of " +
                            "'list[epyt_flow.simulation.scada.SimpleControlModule]' but no of " +
                            f"'{type(simple_controls)}'")
        if len(simple_controls) != 0:
            if any(not isinstance(c, SimpleControlModule) for c in simple_controls):
                raise TypeError("Each item in 'simple_controls' must be an instance of " +
                                "'epyt_flow.simulation.scada.SimppleControlModule'")
        if len(complex_controls) != 0:
            if any(not isinstance(c, ComplexControlModule) for c in complex_controls):
                raise TypeError("Each item in 'complex_controls' must be an instance of " +
                                "'epyt_flow.simulation.scada.ComplexControlModule'")
        if sensor_noise is not None:
            if not isinstance(sensor_noise, SensorNoise):
                raise TypeError("'sensor_noise' must be an instance of " +
                                "'epyt_flow.uncertainty.SensorNoise' but not of " +
                                f"'{type(sensor_noise)}'")
        if model_uncertainty is not None:
            if not isinstance(model_uncertainty, ModelUncertainty):
                raise TypeError("'model_uncertainty' must be an instance of " +
                                "'epyt_flow.uncertainty.ModelUncertainty' but not of " +
                                f"'{type(model_uncertainty)}'")
        if not isinstance(system_events, list):
            raise TypeError("'system_events' must be an instance of " +
                            "'list[epyt_flow.simulation.events.SystemEvent]' but no of " +
                            f"'{type(system_events)}'")
        if len(system_events) != 0:
            if any(not isinstance(c, SystemEvent) for c in system_events):
                raise TypeError("Each item in 'system_events' must be an instance of " +
                                "'epyt_flow.simulation.events.SystemEvent'")
        if not isinstance(sensor_reading_events, list):
            raise TypeError("'sensor_reading_events' must be an instance of " +
                            "'list[epyt_flow.simulation.events.SensorReadingEvent]' but not of " +
                            f"'{type(sensor_reading_events)}'")
        if len(sensor_reading_events) != 0:
            if any(not isinstance(c, SensorReadingEvent) for c in sensor_reading_events):
                raise TypeError("Each item in 'sensor_reading_events' must be an instance of " +
                                "'epyt_flow.simulation.events.SensorReadingEvent'")

        if scenario_config is not None:
            self.__f_inp_in = scenario_config.f_inp_in
            self.__f_msx_in = scenario_config.f_msx_in if f_msx_in is None else f_msx_in

            if general_params is None:
                self.__general_params = scenario_config.general_params
            else:
                self.__general_params = general_params

            if sensor_config is None:
                self.__sensor_config = scenario_config.sensor_config
            else:
                self.__sensor_config = sensor_config

            if memory_consumption_estimate is None:
                self.__memory_consumption_estimate = scenario_config.memory_consumption_estimate
            else:
                self.__memory_consumption_estimate = memory_consumption_estimate

            if len(custom_controls) == 0:
                self.__custom_controls = scenario_config.custom_controls
            else:
                self.__custom_controls = custom_controls

            if len(simple_controls) == 0:
                self.__simple_controls = scenario_config.simple_controls
            else:
                self.__simple_controls = simple_controls

            if len(complex_controls) == 0:
                self.__complex_controls = scenario_config.complex_controls
            else:
                self.__complex_controls = complex_controls

            if sensor_noise is None:
                self.__sensor_noise = scenario_config.sensor_noise
            else:
                self.__sensor_noise = sensor_noise

            if model_uncertainty is None:
                self.__model_uncertainty = scenario_config.model_uncertainty
            else:
                self.__model_uncertainty = model_uncertainty

            if len(system_events) == 0:
                self.__system_events = scenario_config.system_events
            else:
                self.__system_events = system_events

            if len(sensor_reading_events) == 0:
                self.__sensor_reading_events = scenario_config.sensor_reading_events
            else:
                self.__sensor_reading_events = sensor_reading_events
        else:
            self.__f_inp_in = f_inp_in
            self.__f_msx_in = f_msx_in
            self.__general_params = general_params
            self.__sensor_config = sensor_config
            self.__memory_consumption_estimate = memory_consumption_estimate
            self.__custom_controls = custom_controls
            self.__simple_controls = simple_controls
            self.__complex_controls = complex_controls
            self.__sensor_noise = sensor_noise
            self.__system_events = system_events
            self.__sensor_reading_events = sensor_reading_events

            if model_uncertainty is not None:
                self.__model_uncertainty = model_uncertainty
            else:
                self.__model_uncertainty = ModelUncertainty()

        super().__init__(**kwds)

    @property
    def f_inp_in(self) -> str:
        """
        Gets the path to the .inp file.

        Returns
        -------
        `str`
            Path to the .inp file.
        """
        if Path(self.__f_inp_in).is_absolute():
            return self.__f_inp_in
        elif Path(self.__f_inp_in).name == self.__f_inp_in:
            return os.path.join(self._parent_path, self.__f_inp_in)
        else:
            return self.__f_inp_in

    @property
    def f_msx_in(self) -> str:
        """
        Gets the path to the .msx file.

        Returns
        -------
        `str`
            Path to the .msx file.
        """
        if self.__f_msx_in is None:
            return None
        else:
            if Path(self.__f_msx_in).is_absolute():
                return self.__f_msx_in
            elif Path(self.__f_msx_in).name == self.__f_msx_in:
                return os.path.join(self._parent_path, self.__f_msx_in)
            else:
                return self.__f_msx_in

    @property
    def general_params(self) -> dict:
        """
        Gets general parameters such as hydraulic time step, etc.

        Returns
        -------
        `dict`
            All general parameters as dictionary -- the parameter name serves as a key.
        """
        return deepcopy(self.__general_params)

    @property
    def sensor_config(self) -> SensorConfig:
        """
        Gets the sensor configuration.

        Returns
        -------
        :class:`~epyt_flow.simulation.sensor_config.SensorConfig`
            Sensor configuration.
        """
        return deepcopy(self.__sensor_config)

    @property
    def memory_consumption_estimate(self) -> float:
        """
        Gets the estimated memory consumption of this scenario -- i.e. the amount of memory that is
        needed on the hard disk as well as in RAM.

        Returns
        -------
        `float`
            Estimated memory consumption in MB.
        """
        return self.__memory_consumption_estimate

    @property
    def custom_controls(self) -> list[CustomControlModule]:
        """
        Returns the list of all custom control modules that are active during the simulation.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.custom_control.CustomControlModule`]
            List of all custom control modules that are active during the simulation.
        """
        return deepcopy(self.__custom_controls)

    @property
    def simple_controls(self) -> list[SimpleControlModule]:
        """
        Gets the list of all EPANET control rules that are active during the simulation.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.simple_control.SimpleControlModule`]
            List of all EPANET control rules that are active during the simulation.
        """
        return deepcopy(self.__simple_controls)

    @property
    def complex_controls(self) -> list[ComplexControlModule]:
        """
        Gets the list of all complex (i.e. IF-THEN-ELSE) EPANET control rules
        that are active during the simulation.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.complex_control.ComplexControlModule`]
            List of all complex EPANET control rules that are active during the simulation.
        """
        return deepcopy(self.__complex_controls)

    @property
    def sensor_noise(self) -> SensorNoise:
        """
        Gets the sensor noise/uncertainty specification.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.sensor_noise.SensorNoise`
            Sensor noise/uncertainty.
        """
        return deepcopy(self.__sensor_noise)

    @property
    def model_uncertainty(self) -> ModelUncertainty:
        """
        Gets the model uncertainty specification.

        Returns
        -------
        :class:`~epyt_flow.uncertainty.model_uncertainty.ModelUncertainty`
            Model uncertainty specification.
        """
        return deepcopy(self.__model_uncertainty)

    @property
    def system_events(self) -> list[SystemEvent]:
        """
        Gets all system events.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.events.system_event.SystemEvent`]
            All system events.
        """
        return deepcopy(self.__system_events)

    @property
    def sensor_reading_events(self) -> list[SensorReadingEvent]:
        """
        Gets all sensor reading events.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.events.sensor_reading_event.SensorReadingEvent`]
            All sensor reading events.
        """
        return deepcopy(self.__sensor_reading_events)

    def get_attributes(self) -> dict:
        my_attributes = {"f_inp_in": self.__f_inp_in, "f_msx_in": self.__f_msx_in,
                         "general_params": self.__general_params,
                         "sensor_config": self.__sensor_config,
                         "memory_consumption_estimate": self.__memory_consumption_estimate,
                         "custom_controls": self.__custom_controls,
                         "simple_controls": self.__simple_controls,
                         "complex_controls": self.__complex_controls,
                         "sensor_noise": self.__sensor_noise,
                         "model_uncertainty": self.__model_uncertainty,
                         "system_events": self.__system_events,
                         "sensor_reading_events": self.__sensor_reading_events}

        return super().get_attributes() | my_attributes

    def __eq__(self, other) -> bool:
        if not isinstance(other, ScenarioConfig):
            raise TypeError("Can not compare 'ScenarioConfig' instance " +
                            f"with '{type(other)}' instance")

        return self.__f_inp_in == other.f_inp_in and self.__f_msx_in == other.f_msx_in \
            and self.__general_params == other.general_params \
            and self.__memory_consumption_estimate == other.memory_consumption_estimate \
            and self.__sensor_config == other.sensor_config \
            and np.all(self.__custom_controls == other.custom_controls) \
            and np.all(self.__simple_controls == other.simple_controls) \
            and np.all(self.__complex_controls == other.complex_controls) \
            and self.__model_uncertainty == other.model_uncertainty \
            and np.all(self.__system_events == other.system_events) \
            and np.all(self.__sensor_reading_events == other.sensor_reading_events)

    def __str__(self) -> str:
        return f"f_inp_in: {self.f_inp_in} f_msx_in: {self.f_msx_in} " + \
            f"general_params: {self.general_params} sensor_config: {self.sensor_config} " + \
            f"memory_consumption_estimate: {self.memory_consumption_estimate} " + \
            f"simple_controls: {self.simple_controls} " + \
            f"complex_controls: {self.__complex_controls} " + \
            f"custom_controls: {self.__custom_controls}" + \
            f"sensor_noise: {self.sensor_noise} model_uncertainty: {self.model_uncertainty} " + \
            f"system_events: {','.join(map(str, self.system_events))} " + \
            f"sensor_reading_events: {','.join(map(str, self.sensor_reading_events))}"

    @staticmethod
    def load_from_json_file(f_json_in: str) -> Any:
        """
        Loads a scenario configuration from a given JSON file.

        Parameters
        ----------
        f_json_in : `str`
            Path to JSON configuration file.

        Returns
        -------
        :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
            Loaded scenario configuration.
        """
        with open(f_json_in, "r", encoding="utf-8") as f:
            return ScenarioConfig.load_from_json(f.read())

    @staticmethod
    def load_from_json(config_data: str) -> Any:
        """
        Loads a scenario configuration from a given JSON string.

        Parameters
        ----------
        config_data : `str`
            JSON data.

        Returns
        -------
        :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
            Loaded scenario configuration.
        """
        data = json.loads(config_data)

        # General parameters and sensor configuration
        general_settings = data["general"]
        f_inp_in = general_settings["file_inp"]
        f_msx_in = general_settings["file_msx"] if "file_msx" in general_settings.keys() else None

        general_params = {"simulation_duration": general_settings["simulation_duration"],
                          "hydraulic_time_step": general_settings["hydraulic_time_step"],
                          "quality_time_step": general_settings["quality_time_step"]}
        if "reporting_time_step" in general_settings.keys():
            general_params["reporting_time_step"] = general_settings["reporting_time_step"]
        if "reporting_time_start" in general_settings.keys():
            general_params["reporting_time_start"] = general_settings["reporting_time_start"]
        if "demand_model" in general_settings.keys():
            general_params["demand_model"] = general_settings["demand_model"]
        if "quality_model" in general_settings.keys():
            general_params["quality_model"] = general_settings["quality_model"]
        if "flow_units_id" in general_settings.keys():
            general_params["flow_units_id"] = general_settings["flow_units_id"]

        sensor_config = data["sensors"]

        if "pressure_sensors" in sensor_config.keys():
            pressure_sensors = sensor_config["pressure_sensors"]
        else:
            pressure_sensors = []

        if "flow_sensors" in sensor_config.keys():
            flow_sensors = sensor_config["flow_sensors"]
        else:
            flow_sensors = []

        if "demand_sensors" in sensor_config.keys():
            demand_sensors = sensor_config["demand_sensors"]
        else:
            demand_sensors = []

        if "node_quality_sensors" in sensor_config.keys():
            node_quality_sensors = sensor_config["node_quality_sensors"]
        else:
            node_quality_sensors = []

        if "link_quality_sensors" in sensor_config.keys():
            link_quality_sensors = sensor_config["link_quality_sensors"]
        else:
            link_quality_sensors = []

        if "tank_volume_sensors" in sensor_config.keys():
            tank_volume_sensors = sensor_config["tank_volume_sensors"]
        else:
            tank_volume_sensors = []

        if "valve_state_sensors" in sensor_config.keys():
            valve_state_sensors = sensor_config["valve_state_sensors"]
        else:
            valve_state_sensors = []

        if "pump_state_sensors" in sensor_config.keys():
            pump_state_sensors = sensor_config["pump_state_sensors"]
        else:
            pump_state_sensors = []

        if "bulk_species_node_sensors" in sensor_config.keys():
            bulk_species_node_sensors = sensor_config["bulk_species_node_sensors"]
        else:
            bulk_species_node_sensors = {}

        if "bulk_species_link_sensors" in sensor_config.keys():
            bulk_species_link_sensors = sensor_config["bulk_species_link_sensors"]
        else:
            bulk_species_link_sensors = {}

        if "surface_species_sensors" in sensor_config.keys():
            surface_species_sensors = sensor_config["surface_species_sensors"]
        else:
            surface_species_sensors = {}

        # Uncertainties
        if "uncertainties" in data.keys():
            def parse_uncertantiy(uncertainty_desc: dict) -> Uncertainty:
                uncertainty_type = uncertainty_desc["type"]
                del uncertainty_desc["type"]

                if uncertainty_type == "absolute_gaussian":
                    return AbsoluteGaussianUncertainty(**uncertainty_desc)
                elif uncertainty_type == "relative_gaussian":
                    return RelativeGaussianUncertainty(**uncertainty_desc)
                elif uncertainty_type == "absolute_uniform":
                    return AbsoluteUniformUncertainty(**uncertainty_desc)
                elif uncertainty_type == "relative_uniform":
                    return RelativeUniformUncertainty(**uncertainty_desc)
                else:
                    raise ValueError(f"Unknown uncertainty '{uncertainty_type}'")

            uncertanties = data["uncertainties"]
            if "pipe_length_uncertainty" in uncertanties.keys():
                pipe_length_uncertainty = parse_uncertantiy(uncertanties["pipe_length_uncertainty"])
            else:
                pipe_length_uncertainty = None
            if "pipe_roughness_uncertainty" in uncertanties.keys():
                pipe_roughness_uncertainty = parse_uncertantiy(
                    uncertanties["pipe_roughness_uncertainty"])
            else:
                pipe_roughness_uncertainty = None
            if "pipe_diameter_uncertainty" in uncertanties.keys():
                pipe_diameter_uncertainty = parse_uncertantiy(
                    uncertanties["pipe_diameter_uncertainty"])
            else:
                pipe_diameter_uncertainty = None
            if "demand_base_uncertainty" in uncertanties.keys():
                demand_base_uncertainty = parse_uncertantiy(uncertanties["demand_base_uncertainty"])
            else:
                demand_base_uncertainty = None
            if "demand_pattern_uncertainty" in uncertanties.keys():
                demand_pattern_uncertainty = parse_uncertantiy(
                    uncertanties["demand_pattern_uncertainty"])
            else:
                demand_pattern_uncertainty = None
            if "elevation_uncertainty" in uncertanties.keys():
                elevation_uncertainty = parse_uncertantiy(uncertanties["elevation_uncertainty"])
            else:
                elevation_uncertainty = None
            if "constants_uncertainty" in uncertanties.keys():
                constants_uncertainty = parse_uncertantiy(uncertanties["constants_uncertainty"])
            else:
                constants_uncertainty = None
            if "parameters_uncertainty" in uncertanties.keys():
                parameters_uncertainty = parse_uncertantiy(uncertanties["parameters_uncertainty"])
            else:
                parameters_uncertainty = None

            model_uncertainty = ModelUncertainty(pipe_length_uncertainty,
                                                 pipe_roughness_uncertainty,
                                                 pipe_diameter_uncertainty, demand_base_uncertainty,
                                                 demand_pattern_uncertainty, elevation_uncertainty,
                                                 constants_uncertainty, parameters_uncertainty)

            if "sensor_noise" in uncertanties.keys():
                sensor_noise = SensorNoise(parse_uncertantiy(uncertanties["sensor_noise"]))
            else:
                sensor_noise = None

        # Events
        leakages = []
        if "leakages" in data.keys():
            def parse_leak(leak_desc):
                leak_type = leak_desc["type"]
                del leak_desc["type"]

                if leak_type == "abrupt":
                    return AbruptLeakage(**leak_desc)
                elif leak_type == "incipient":
                    return IncipientLeakage(**leak_desc)
                else:
                    raise ValueError(f"Unknown leakage type '{leak_type}'")

            leakages = [parse_leak(leak) for leak in data["leakages"]]

        sensor_faults = []
        if "sensor_faults" in data.keys():
            def parse_sensor_fault(sensor_fault_desc):
                fault_type = sensor_fault_desc["type"]
                del sensor_fault_desc["type"]

                if fault_type == "constant":
                    return SensorFaultConstant(**sensor_fault_desc)
                elif fault_type == "drift":
                    return SensorFaultDrift(**sensor_fault_desc)
                elif fault_type == "gaussian":
                    return SensorFaultGaussian(**sensor_fault_desc)
                elif fault_type == "percentage":
                    return SensorFaultPercentage(**sensor_fault_desc)
                elif fault_type == "stuckatzero":
                    return SensorFaultStuckZero(**sensor_fault_desc)
                else:
                    raise ValueError(f"Unknown sensor fault '{fault_type}'")

            sensor_faults = [parse_sensor_fault(sensor_fault)
                             for sensor_fault in data["sensor_faults"]]

        #  Load .inp file to get a list of all nodes and links/pipes
        sensor_config = None
        from .scenario_simulator import ScenarioSimulator
        with ScenarioSimulator(f_inp_in) as scenario:
            sensor_config = SensorConfig.create_empty_sensor_config(scenario.sensor_config)
            sensor_config.pressure_sensors = pressure_sensors
            sensor_config.flow_sensors = flow_sensors
            sensor_config.demand_sensors = demand_sensors
            sensor_config.quality_node_sensors = node_quality_sensors
            sensor_config.quality_link_sensors = link_quality_sensors
            sensor_config.valve_state_sensors = valve_state_sensors
            sensor_config.pump_state_sensors = pump_state_sensors
            sensor_config.tank_volume_sensors = tank_volume_sensors
            sensor_config.bulk_species_node_sensors = bulk_species_node_sensors
            sensor_config.bulk_species_link_sensors = bulk_species_link_sensors
            sensor_config.surface_species_sensors = surface_species_sensors

        # Create final scenario configuration
        return ScenarioConfig(f_inp_in=f_inp_in, f_msx_in=f_msx_in, general_params=general_params,
                              sensor_config=sensor_config, sensor_noise=sensor_noise,
                              model_uncertainty=model_uncertainty, system_events=leakages,
                              sensor_reading_events=sensor_faults)



"""
This module provides a base handler with some useful methods for all REST API handlers.
"""
from typing import Any
import falcon

from ..serialization import my_load_from_json, my_to_json


class BaseHandler():
    """
    Base class for all REST API handlers.
    """
    def send_error(self, resp: falcon.Response, error_msg: str) -> None:
        """
        Sends an error message back.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        error_msg : `str`
            Error message.
        """
        resp.status = falcon.HTTP_BAD_REQUEST
        resp.data = error_msg.encode()

    def send_invalid_resource_id_error(self, resp: falcon.Response) -> None:
        """
        Sends an error that th given resource ID (e.g. scenario ID, or SCADA data ID) is invalid.

        Parameters
        ----------
        resp : `falcon.Response`
            Response instance.
        """
        resp.status = falcon.HTTP_BAD_REQUEST
        resp.data = "Invalid resource ID".encode()

    def send_json_parsing_error(self, resp: falcon.Response) -> None:
        """
        Sends an error that the JSON parsing failed.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        """
        resp.status = falcon.HTTP_BAD_REQUEST
        resp.data = "Failed to parse JSON".encode()

    def load_json_data_from_request(self, req: falcon.Request) -> Any:
        """
        Loads/Parses an object from given JSON data.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.

        Returns
        -------
        `Any`
            Loaded object.
        """
        try:
            return my_load_from_json(req.bounded_stream.read())
        except Exception:
            return None

    def send_json_response(self, resp: falcon.Response, data: Any) -> None:
        """
        Sends a JSON response.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        data : `Any`
            Data to be sent.
        """
        resp.content_type = falcon.MEDIA_JSON
        resp.status = falcon.HTTP_200
        resp.data = my_to_json(data).encode()



"""
Module provides base classes for system events such as leakages, actuator events, etc.
"""
from abc import abstractmethod
from epanet_plus import EPyT

from .event import Event


class SystemEvent(Event):
    """
    Base class for a system event -- i.e. an event that affects the EPANET simulation.
    """
    def __init__(self, **kwds):
        self._epanet_api = None
        self._exit_called = False

        super().__init__(**kwds)

    def init(self, epanet_api: EPyT) -> None:
        """
        Initializes the event.

        Parameters
        ----------
        epanet_api : `epanet_plus.EPyT <https://epanet-plus.readthedocs.io/en/stable/api.html#epanet_plus.epanet_toolkit.EPyT>`_
            API to EPANET and EPANET-MSX.
        """
        self._epanet_api = epanet_api

    def __getstate__(self):
        state = self.__dict__.copy()
        if "_epanet_api" in state:
            # Do not serialize EPyT instance!
            del state["_epanet_api"]

        return state

    def __call__(self, cur_time) -> None:
        return self.step(cur_time)

    def step(self, cur_time) -> None:
        """
        Is called at every iteration (time step) in the simulation.
        `apply` or `exit` are called if necessary.

        Parameters
        ----------
        cur_time : `int`
            Current time (seconds since the start) in the simulation.
        """
        if self.start_time <= cur_time < self.end_time:
            self.apply(cur_time)
        elif cur_time > self.end_time:
            if self._exit_called is False:
                self.exit(cur_time)
                self._exit_called = True

    def reset(self) -> None:
        """
        Resets this event -- i.e. make it ready for another simulation run.
        """

    def exit(self, cur_time) -> None:
        """
        Is called ONCE after the event is over -- i.e. next time step after `end_time`.

        Any "clean-up" or "exiting" logic should go here.

        Parameters
        ----------
        cur_time : `int`
            Current time (seconds since the start) in the simulation.
        """

    def cleanup(self) -> None:
        """
        Clean up any changes/modifications made by this event.
        """

    @abstractmethod
    def apply(self, cur_time: int) -> None:
        """
        Implements the event using EPANET and EPANET-MSX.

        This function is only called when the event is active.

        Parameters
        ----------
        cur_time : `int`
            Current time (seconds since the start) in the simulation.
        """
        raise NotImplementedError()



from .battledim import load_scenario as load_battledim_scenario, \
    load_scada_data as load_battledim_scada_data, load_data as load_battledim_data, \
    compute_evaluation_score as compute_battledim_evaluation_score
from .leakdb import load_scenarios as load_leakdb_scenarios, \
    load_scada_data as load_leakdb_scada_data, load_data as load_leakdb_data, \
    compute_evaluation_score as compute_leakdb_evaluation_score
from .batadal import load_scenario as load_batadal_scenario, \
    load_scada_data as load_batadal_scada_data, load_data as load_batadal_data
from .gecco_water_quality import load_gecco2017_water_quality_data, \
    load_gecco2018_water_quality_data, load_gecco2019_water_quality_data
from .water_usage import load_water_usage



from .system_event import *
from .leakages import *
from .quality_events import *
from .sensor_reading_event import *
from .sensor_faults import *
from .sensor_reading_attack import *
from .actuator_events import *



"""
Module provides classes for implementing different sensor faults.
"""
from abc import abstractmethod
import numpy as np

from .sensor_reading_event import SensorReadingEvent
from ...serialization import serializable, JsonSerializable, SENSOR_FAULT_CONSTANT_ID, \
    SENSOR_FAULT_DRIFT_ID, SENSOR_FAULT_GAUSSIAN_ID, SENSOR_FAULT_PERCENTAGE_ID, \
    SENSOR_FAULT_STUCKATZERO_ID


class SensorFault(SensorReadingEvent):
    """
    Base class for a sensor fault
    """
    # Acknowledgement: This Python implementation is based on
    # https://github.com/eldemet/sensorfaultmodels/blob/main/sensorfaultmodels.m
    # and https://github.com/Mariosmsk/sensorfaultmodels/blob/main/sensorfaultmodels.py

    def compute_multiplier(self, cur_time: int) -> float:
        """
        Computes the multiplier for a given time stamp.

        Parameters
        ----------
        cur_time : `int`
            Time in seconds.

        Returns
        -------
        `float`
            Multiplier.
        """
        b1 = 0
        b2 = 0
        a1 = 1
        a2 = 1

        if cur_time >= self.start_time:
            b1 = 1 - np.exp(- a1 * (cur_time - self.start_time))

        if cur_time >= self.end_time:
            b2 = 1 - np.exp(- a2 * (cur_time - self.end_time))

        return b1 - b2

    @abstractmethod
    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        """
        Applies this sensor fault to a given single sensor reading value --
        i.e. the sensor reading value is perturbed by this fault.

        Parameters:
        -----------
        cur_multiplier : `float`
            Current multiplier -- i.e. controls the "strength" of the fault.
        sensor_reading : `float`
            Sensor reading value.
        cur_time : `int`
            Current time stamp (in seconds) in the simulation.

        Returns
        -------
        `float`
            Perturbed sensor reading value.
        """
        raise NotImplementedError()

    def apply(self, sensor_readings: np.ndarray,
              sensor_readings_time: np.ndarray) -> np.ndarray:
        for i in range(sensor_readings.shape[0]):
            t = sensor_readings_time[i]
            sensor_readings[i] = self.apply_sensor_fault(self.compute_multiplier(t),
                                                         sensor_readings[i], t)

        return sensor_readings


@serializable(SENSOR_FAULT_CONSTANT_ID, ".epytflow_sensorfault_constant")
class SensorFaultConstant(SensorFault, JsonSerializable):
    """
    Class implementing a constant shift sensor fault.

    Parameters
    ----------
    constant_shift : `float`
        Constant that is added to the sensor reading.
    """
    def __init__(self, constant_shift: float, **kwds):
        if not isinstance(constant_shift, float):
            raise TypeError("'constant_shift' must be an instance of 'float' but no of " +
                            f"'{type(constant_shift)}'")

        self._constant_shift = constant_shift

        super().__init__(**kwds)

    @property
    def constant_shift(self) -> float:
        """
        Gets the Constant that is added to the sensor reading.

        Returns
        -------
        `float`
            Constant that is added to the sensor reading.
        """
        return self._constant_shift

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"constant_shift": self._constant_shift}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._constant_shift == other.constant_shift

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} constant: {self._constant_shift}"

    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        return sensor_reading + cur_multiplier * self._constant_shift


@serializable(SENSOR_FAULT_DRIFT_ID, ".epytflow_sensorfault_drift")
class SensorFaultDrift(SensorFault, JsonSerializable):
    """
    Class implementing a drift sensor fault.

    Parameters
    ----------
    coef : `float`
        Coefficient of the drift.
    """
    def __init__(self, coef: float, **kwds):
        self._coef = coef

        super().__init__(**kwds)

    @property
    def coef(self) -> float:
        """
        Gets the coefficient of the drift.

        Returns
        -------
        `float`
            Coefficient of the drift.
        """
        return self._coef

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"coef": self._coef}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._coef == other.coef

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} coef: {self._coef}"

    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        return sensor_reading + cur_multiplier * (self._coef * (cur_time - self.start_time))


@serializable(SENSOR_FAULT_GAUSSIAN_ID, ".epytflow_sensorfault_gaussian")
class SensorFaultGaussian(SensorFault, JsonSerializable):
    """
    Class implementing a Gaussian shift sensor fault -- i.e.
     adding Gaussian noise (centered at zero) to the sensor reading.

    Parameters
    ----------
    std : `float`
        Standard deviation of the Gaussian noise.
    """
    def __init__(self, std: float, **kwds):
        if not isinstance(std, float) or not std > 0:
            raise ValueError("'std' must be an instance of 'float' and be greater than 0")

        self._std = std

        super().__init__(**kwds)

    @property
    def std(self) -> float:
        """
        Gets the standard deviation of the Gaussian noise.

        Returns
        -------
        `float`
            Standard deviation of the Gaussian noise.
        """
        return self._std

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"std": self._std}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._std == other.std

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} std: {self._std}"

    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        return sensor_reading + cur_multiplier * np.random.normal(loc=0, scale=self._std)


@serializable(SENSOR_FAULT_PERCENTAGE_ID, ".epytflow_sensorfault_percentage",)
class SensorFaultPercentage(SensorFault, JsonSerializable):
    """
    Class implementing a percentage shift sensor fault.

    Parameters
    ----------
    coef : `float`
        Coefficient (percentage) of the shift -- i.e. coef must be in (0,].
    """
    def __init__(self, coef: float, **kwds):
        if not isinstance(coef, float) or not coef > 0:
            raise ValueError("'coef' must be an instance of 'float' and be greater than zero.")

        self._coef = coef

        super().__init__(**kwds)

    @property
    def coef(self) -> float:
        """
        Gets the coefficient (percentage) of the shift.

        Returns
        -------
        `float`
            Coefficient (percentage) of the shift.
        """
        return self._coef

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"coef": self._coef}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self._coef == other.coef

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()} coef: {self._coef}"

    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        return sensor_reading + cur_multiplier * self._coef * sensor_reading


@serializable(SENSOR_FAULT_STUCKATZERO_ID, ".epytflow_sensorfault_zero")
class SensorFaultStuckZero(SensorFault, JsonSerializable):
    """
    Class implementing a stuck-at-zero sensor fault -- i.e. sensor reading is set to zero.
    """

    def __str__(self) -> str:
        return f"{type(self).__name__} {super().__str__()}"

    def apply_sensor_fault(self, cur_multiplier: float, sensor_reading: float,
                           cur_time: int) -> float:
        return sensor_reading + cur_multiplier * (-1. * sensor_reading)



"""
Module contains a class for representing complex control rules as implemented in EPANET.
"""
from copy import deepcopy
from typing import Any
import numpy as np
from epanet_plus import EpanetConstants

from ...serialization import JsonSerializable, COMPLEX_CONTROL_ID, COMPLEX_CONTROL_CONDITION_ID, \
    COMPLEX_CONTROL_ACTION_ID, serializable


EN_R_AND = 2
EN_R_OR  = 3

EN_R_DEMAND    = 1
EN_R_HEAD      = 2
EN_R_LEVEL     = 3
EN_R_PRESSURE  = 4
EN_R_FLOW      = 5
EN_R_STATUS    = 6
EN_R_SETTING   = 7
EN_R_POWER     = 8
EN_R_TIME      = 9
EN_R_CLOCKTIME = 10
EN_R_FILLTIME  = 11
EN_R_DRAINTIME = 12

EN_R_EQ      = 0
EN_R_NEQ     = 1
EN_R_LEQ     = 2
EN_R_GEQ     = 3
EN_R_LESS    = 4
EN_R_GREATER = 5
EN_R_IS      = 6
EN_R_NOT     = 7
EN_R_BELOW   = 8
EN_R_ABOVE   = 9

EN_R_ACTION_SETTING       = -1
EN_R_ACTION_STATUS_OPEN   = 1
EN_R_ACTION_STATUS_CLOSED = 2
EN_R_ACTION_STATUS_ACTIVE = 3

RULESTATUS = ['OPEN', 'CLOSED', 'ACTIVE']


@serializable(COMPLEX_CONTROL_CONDITION_ID, ".epytflow_complex_control_condition")
class RuleCondition(JsonSerializable):
    """
    Class representing a rule condition.

    Parameters
    ----------
    object_type_id : `int`
        ID of the object type.

        Must be one of the following EPANET constants:

            - EN_R_NODE   = 6
            - EN_R_LINK   = 7
            - EN_R_SYSTEM = 8
    object_id : `str`
        ID of the object (i.e. junction, pipe, link, tank, etc.).
    attribute_id : `int`
        Type ID of the object's attribute that is checked.

        Must be on of the following constants:

            - EN_R_DEMAND    = 1
            - EN_R_HEAD      = 2
            - EN_R_LEVEL     = 3
            - EN_R_PRESSURE  = 4
            - EN_R_FLOW      = 5
            - EN_R_STATUS    = 6
            - EN_R_SETTING   = 7
            - EN_R_TIME      = 9
            - EN_R_CLOCKTIME = 10
            - EN_R_FILLTIME  = 11
            - EN_R_DRAINTIME = 12
    relation_type_id : `int`
        ID of the type of comparison.

        Must be one of the following constants:

            - EN_R_EQ      = 0
            - EN_R_NEQ     = 1
            - EN_R_LEQ     = 2
            - EN_R_GEQ     = 3
            - EN_R_LESS    = 4
            - EN_R_GREATER = 5
            - EN_R_IS      = 6
            - EN_R_NOT     = 7
            - EN_R_BELOW   = 8
            - EN_R_ABOVE   = 9
    value : `Any`
        Value that is compared against.
    """
    def __init__(self, object_type_id: int, object_id: str, attribute_id: int,
                 relation_type_id: int, value: Any, **kwds):
        if not isinstance(object_type_id, int):
            raise TypeError("'object_type_id' must be an instance of 'int' " +
                            f"but not of '{type(object_type_id)}'")
        if object_type_id not in [EpanetConstants.EN_R_NODE, EpanetConstants.EN_R_LINK,
                                  EpanetConstants.EN_R_SYSTEM]:
            raise ValueError(f"Invalid value '{object_type_id}' for 'object_type_id'")
        if not isinstance(object_id, str):
            raise TypeError("'object_id' must be an instance of 'str' " +
                            f"but not of '{type(object_id)}'")
        if not isinstance(attribute_id, int):
            raise TypeError("'attribute_id' must be an instance of 'int' " +
                            f"but not of '{type(attribute_id)}'")
        if attribute_id not in [EN_R_DEMAND, EN_R_HEAD, EN_R_LEVEL, EN_R_PRESSURE,
                                EN_R_FLOW, EN_R_STATUS, EN_R_SETTING, EN_R_POWER, EN_R_TIME,
                                EN_R_CLOCKTIME, EN_R_FILLTIME, EN_R_DRAINTIME]:
            raise ValueError(f"Invalid value '{attribute_id}' for 'attribute_id'")
        if not isinstance(relation_type_id, int):
            raise TypeError("'relation_type_id' must be an instance of 'int' " +
                            f"but not of '{type(relation_type_id)}'")
        if relation_type_id not in [EN_R_EQ, EN_R_NEQ, EN_R_LEQ, EN_R_GEQ, EN_R_LESS, EN_R_GREATER,
                                    EN_R_IS, EN_R_NOT, EN_R_BELOW, EN_R_ABOVE]:
            raise ValueError(f"Invalid value '{relation_type_id}' for 'relation_type_id'")

        self.__object_type_id = object_type_id
        self.__object_id = object_id
        self.__attribute_id = attribute_id
        self.__relation_type_id = relation_type_id
        self.__value = value

        super().__init__(**kwds)

    @property
    def object_type_id(self) -> int:
        """
        Returns the ID of the object type.

        Will be one of the following EPANET constants:

            - EN_R_NODE   = 6
            - EN_R_LINK   = 7
            - EN_R_SYSTEM = 8

        Returns
        -------
        `int`
            ID of the object type..
        """
        return self.__object_type_id

    @property
    def object_id(self) -> str:
        """
        Returns the ID of the object (i.e. junction, pipe, link, tank, etc.).

        Returns
        -------
        `str`
            ID of the object.
        """
        return self.__object_id

    @property
    def attribute_id(self) -> int:
        """
        Returns the type ID of the object's attribute that is checked.

        Will be one of the following constants:

            - EN_R_DEMAND    = 1
            - EN_R_HEAD      = 2
            - EN_R_LEVEL     = 3
            - EN_R_PRESSURE  = 4
            - EN_R_FLOW      = 5
            - EN_R_STATUS    = 6
            - EN_R_SETTING   = 7
            - EN_R_TIME      = 9
            - EN_R_CLOCKTIME = 10
            - EN_R_FILLTIME  = 11
            - EN_R_DRAINTIME = 12

        Returns
        -------
        `int`
            Type ID of the object's attribute that is checked.
        """
        return self.__attribute_id

    @property
    def relation_type_id(self) -> int:
        """
        Returns the ID of the type of comparison.

        Will be one of the following constants:

            - EN_R_EQ      = 0
            - EN_R_NEQ     = 1
            - EN_R_LEQ     = 2
            - EN_R_GEQ     = 3
            - EN_R_LESS    = 4
            - EN_R_GREATER = 5
            - EN_R_IS      = 6
            - EN_R_NOT     = 7
            - EN_R_BELOW   = 8
            - EN_R_ABOVE   = 9

        Returns
        -------
        `int`
            ID of the type of comparison.
        """
        return self.__relation_type_id

    @property
    def value(self) -> Any:
        """
        Returns the value that is compared against.

        Returns
        -------
        `Any`
            Value that is compared against.
        """
        return self.__value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"object_type_id": self.__object_type_id,
                                           "object_id": self.__object_id,
                                           "attribute_id": self.__attribute_id,
                                           "relation_type_id": self.__relation_type_id,
                                           "value": self.__value}

    def __eq__(self, other) -> bool:
        return self.__object_type_id == other.object_type_id and \
            self.__object_id == other.object_id and self.__attribute_id == other.attribute_id and \
            self.__relation_type_id == other.relation_type_id and self.__value == other.value

    def __str__(self) -> str:
        desc = ""

        if self.__attribute_id == EN_R_DEMAND:
            if self.__object_type_id == EpanetConstants.EN_R_NODE:
                desc += f"JUNCTION {self.__object_id} DEMAND "
            elif self.__object_type_id == EpanetConstants.EN_R_SYSTEM:
                desc += "SYSTEM DEMAND "
        elif self.__attribute_id == EN_R_HEAD:
            desc += f"JUNCTION {self.__object_id} HEAD "
        elif self.__attribute_id == EN_R_LEVEL:
            desc += f"TANK {self.__object_id} LEVEL "
        elif self.__attribute_id == EN_R_PRESSURE:
            desc += f"JUNCTION {self.__object_id} PRESSURE "
        elif self.__attribute_id == EN_R_FLOW:
            desc += f"LINK {self.__object_id} FLOW "
        elif self.__attribute_id == EN_R_STATUS:
            desc += f"LINK {self.__object_id} STATUS "
        elif self.__attribute_id == EN_R_SETTING:
            desc += f"LINK {self.__object_id} SETTING "
        elif self.__attribute_id == EN_R_TIME:
            desc += "SYSTEM TIME "
        elif self.__attribute_id == EN_R_CLOCKTIME:
            desc += "SYSTEM CLOCKTIME "
        elif self.__attribute_id == EN_R_FILLTIME:
            desc += f"TANK {self.__object_id} FILLTIME "
        elif self.__attribute_id == EN_R_DRAINTIME:
            desc += f"TANK {self.__object_id} DRAINTIME "

        if self.__relation_type_id == EN_R_EQ:
            desc += "= "
        elif self.__relation_type_id == EN_R_IS:
            desc += "IS "
        elif self.__relation_type_id == EN_R_NOT:
            desc += "IS NOT "
        elif self.__relation_type_id == EN_R_LEQ:
            desc += "<= "
        elif self.__relation_type_id == EN_R_GEQ:
            desc += ">= "
        elif self.__relation_type_id == EN_R_ABOVE:
            desc += "ABOVE "
        elif self.__relation_type_id == EN_R_BELOW:
            desc += "BELOW "
        elif self.__relation_type_id == EN_R_LESS:
            desc += "< "
        elif self.__relation_type_id == EN_R_GREATER:
            desc += "> "

        desc += str(self.__value)

        return desc


@serializable(COMPLEX_CONTROL_ACTION_ID, ".epytflow_complex_control_action")
class RuleAction(JsonSerializable):
    """
    Class representing a rule action.

    Parameters
    ----------
    link_type_id : `int`
        Link type ID.

        Must be one of following EPANET constants:

            - EN_CVPIPE = 0
            - EN_PIPE = 1
            - EN_PUMP = 2
            - EN_PRV = 3
            - EN_PSV = 4
            - EN_PBV = 5
            - EN_FCV = 6
            - EN_TCV = 7
            - EN_GPV = 8
    link_id : `str`
        Link ID.
    action_type_id : `int`
        Type ID of the action.

        Must be one of the following constants:

            - EN_R_ACTION_SETTING       = -1
            - EN_R_ACTION_STATUS_OPEN   = 1
            - EN_R_ACTION_STATUS_CLOSED = 2
            - EN_R_ACTION_STATUS_ACTIVE = 3
    action_value : `Any`
        Value of the acton (e.g. pump speed).
        Only relevant if action_type_id = EN_R_SETTING, will be ignored in all other cases.
    """
    def __init__(self, link_type_id: int, link_id: str, action_type_id: int, action_value: Any,
                 **kwds):
        if not isinstance(link_type_id, int):
            raise TypeError("'link_type_id' must be an istanace of 'int' " +
                            f"but not of '{type(link_type_id)}'")
        if link_type_id not in [EpanetConstants.EN_CVPIPE, EpanetConstants.EN_PIPE,
                                EpanetConstants.EN_PUMP, EpanetConstants.EN_PRV,
                                EpanetConstants.EN_PSV, EpanetConstants.EN_PBV,
                                EpanetConstants.EN_FCV, EpanetConstants.EN_TCV,
                                EpanetConstants.EN_GPV]:
            raise ValueError(f"Invalid value '{link_type_id}' for 'link_type_id'")
        if not isinstance(link_id, str):
            raise TypeError("'link_id' must be an instance of 'str' " +
                            f"but not of '{type(link_id)}'")
        if not isinstance(action_type_id, int):
            raise TypeError("'action_type_id' must be an instance of 'int' " +
                            f"but not of '{type(action_type_id)}'")
        if action_type_id not in [EN_R_ACTION_SETTING, EN_R_ACTION_STATUS_OPEN,
                                  EN_R_ACTION_STATUS_CLOSED, EN_R_ACTION_STATUS_ACTIVE]:
            raise ValueError(f"Invalid value '{action_type_id}' for 'action_type_id'")

        self.__link_type_id = link_type_id
        self.__link_id = link_id
        self.__action_type_id = action_type_id
        self.__action_value = action_value

        super().__init__(**kwds)

    @property
    def link_type_id(self) -> int:
        """
        Returns the link type ID.

        Will be one of the following EPANET constants:

            - EN_CVPIPE = 0
            - EN_PIPE = 1
            - EN_PUMP = 2
            - EN_PRV = 3
            - EN_PSV = 4
            - EN_PBV = 5
            - EN_FCV = 6
            - EN_TCV = 7
            - EN_GPV = 8

        Returns
        -------
        `int`
            Link type ID.
        """
        return self.__link_type_id

    @property
    def link_id(self) -> str:
        """
        Returns the link ID.

        Returns
        -------
        `str`
            Link ID.
        """
        return self.__link_id

    @property
    def action_type_id(self) -> int:
        """
        Returns the type ID of the action.

        Will be one of the following constants:

            - EN_R_ACTION_SETTING       = -1
            - EN_R_ACTION_STATUS_OPEN   = 1
            - EN_R_ACTION_STATUS_CLOSED = 2
            - EN_R_ACTION_STATUS_ACTIVE = 3

        Returns
        -------
        `int`
            Type ID of the action.
        """
        return self.__action_type_id

    @property
    def action_value(self) -> Any:
        """
        Returns the value of the acton (e.g. pump speed).
        Only relevant if action_type_id = EN_R_SETTING.

        Returns
        -------
        `Any`
            Value of the action.
        """
        return self.__action_value

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"link_type_id": self.__link_type_id,
                                           "link_id": self.__link_id,
                                           "action_type_id": self.__action_type_id,
                                           "action_value": self.__action_value}

    def __eq__(self, other) -> bool:
        return self.__link_type_id == other.link_type_id and \
            self.__link_id == other.link_id and \
            self.__action_type_id == other.action_type_id and \
            self.__action_value == other.action_value

    def __str__(self) -> str:
        desc = ""

        if self.__link_type_id in [EpanetConstants.EN_CVPIPE, EpanetConstants.EN_PIPE]:
            desc += "PIPE "
        elif self.__link_type_id == EpanetConstants.EN_PUMP:
            desc += "PUMP "
        else:
            desc += "VALVE "

        desc += f"{self.__link_id} "

        if self.__action_type_id == EN_R_ACTION_SETTING:
            desc += f"SETTING IS {self.__action_value}"
        elif self.__action_type_id == EN_R_ACTION_STATUS_OPEN:
            desc += "STATUS IS OPEN"
        elif self.__action_type_id == EN_R_ACTION_STATUS_CLOSED:
            desc += "STATUS IS CLOSED"
        elif self.action_type_id == EN_R_ACTION_STATUS_ACTIVE:
            desc += "STATUS IS ACTIVE"

        return desc


@serializable(COMPLEX_CONTROL_ID, ".epytflow_complex_control")
class ComplexControlModule(JsonSerializable):
    """
    Class representing a complex control module (i.e. IF-THEN-ELSE rule) as implemented in EPANET.

    Parameters
    ----------
    rule_id : `str`
        ID of the rule.
    condition_1 : :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`
        First condition of this rule.
    additional_conditions : list[tuple[int, :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`]]
        List of (optional) additional conditions incl. their conjunction operator
        (must be either EN_R_AND = 2 or EN_R_OR  = 3).

        Empty list if there are no additional conditions.
    actions : list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
        List of actions that are applied if the conditions are met.
        Must contain at least one action.
    else_actions : list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
        List of actions that are applied if the conditions are NOT met.
    priority : `int`
        Priority of this control rule.
    """
    def __init__(self, rule_id: str, condition_1: RuleCondition,
                 additional_conditions: list[tuple[int, RuleCondition]], actions: list[RuleAction],
                 else_actions: list[RuleAction], priority: int, **kwds):
        if not isinstance(rule_id, str):
            raise TypeError(f"'rule_id' must be an instance of 'str' but not of '{type(rule_id)}'")
        if not isinstance(condition_1, RuleCondition):
            raise TypeError("'condition_1' must be an instance of " +
                            "'epyt_flow.simulation.scada.RuleCondition' " +
                            f"but not of '{type(condition_1)}'")
        if not isinstance(additional_conditions, list) or \
                any(not isinstance(condition, tuple) for condition in additional_conditions) or \
                any(not isinstance(condition[0], int) or condition[0] not in [EN_R_AND, EN_R_OR] or
                    not isinstance(condition[1], RuleCondition)
                    for condition in additional_conditions):
            raise TypeError("'additional_conditions' must be a list of " +
                            "'tuple[int, epyt_flow.simulation.scada.RuleCondition]' instances")
        if not isinstance(actions, list) or any(not isinstance(action, RuleAction)
                                                for action in actions):
            raise TypeError("'actions' must be a list of " +
                            "'epyt_flow.simulation.scada.RuleAction' instances")
        if len(actions) == 0:
            raise ValueError("'actions' must contain at least one action")
        if not isinstance(else_actions, list) or any(not isinstance(action, RuleAction)
                                                     for action in actions):
            raise TypeError("'else_actions' must be a list of " +
                            "'epyt_flow.simulation.scada.RuleAction' instances")
        if not isinstance(priority, int) or priority < 0:
            raise TypeError("'priority' must be a non-negative integer")

        self.__rule_id = rule_id
        self.__condition_1 = condition_1
        self.__additional_conditions = additional_conditions
        self.__actions = actions
        self.__else_actions = else_actions
        self.__priority = priority

        super().__init__(**kwds)

    @property
    def rule_id(self) -> str:
        """
        Returns the ID of this control rule.

        Returns
        -------
        `str`
            ID of this control rule.
        """
        return self.__rule_id

    @property
    def condition_1(self) -> RuleCondition:
        """
        Returns the first condition of this rule.

        Returns
        -------
        :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`
            First condition of this rule.
        """
        return deepcopy(self.__condition_1)

    @property
    def additional_conditions(self) -> list[tuple[int, RuleCondition]]:
        """
        Returns the list of (optional) additional conditions incl. their conjunction operator.
        Empty list if there are no additional conditions.

        Returns
        -------
        list[tuple[int, :class:`~epyt_flow.simulation.scada.complex_control.RuleCondition`]]
            List of (optional) additional conditions incl. their conjunction operator.
        """
        return deepcopy(self.__additional_conditions)

    @property
    def actions(self) -> list[RuleAction]:
        """
        Returns the list of actions that are applied if the conditions are met.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
            List of actions that are applied if the conditions are met.
        """
        return deepcopy(self.__actions)

    @property
    def else_actions(self) -> list[RuleAction]:
        """
        Returns the list of actions that are applied if the conditions are NOT met.

        Returns
        -------
        list[:class:`~epyt_flow.simulation.scada.complex_control.RuleAction`]
            List of actions that are applied if the conditions are NOT met.
        """
        return deepcopy(self.__else_actions)

    @property
    def priority(self) -> int:
        """
        Returns the priority of this control rule.

        Returns
        -------
        `int`
            Priority of this control rule.
        """
        return self.__priority

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"rule_id": self.__rule_id,
                                           "condition_1": self.__condition_1,
                                           "additional_conditions": self.__additional_conditions,
                                           "actions": self.__actions,
                                           "else_actions": self.__else_actions,
                                           "priority": self.__priority}

    def __eq__(self, other) -> bool:
        return super().__eq__(other) and self.__rule_id == other.rule_id and \
            self.__priority == other.priority and self.__condition_1 == other.condition_1 and \
            np.all(self.__additional_conditions == other.additional_conditions) and \
            np.all(self.__actions == other.actions) and \
            np.all(self.__else_actions == other.else_actions)

    def __str__(self) -> str:
        desc = ""

        desc += f"RULE {self.__rule_id}\n"
        desc += f"IF {self.__condition_1} "
        for op, action in self.__additional_conditions:
            if op == EN_R_AND:
                desc += "\nAND "
            elif op == EN_R_OR:
                desc += "\nOR "

            desc += f"{action} "
        desc += "\nTHEN "
        desc += "\nAND ".join(str(action) for action in self.__actions)
        if len(self.__else_actions) != 0:
            desc += "\nELSE " + "\nAND ".join(str(action) for action in self.__else_actions)

        desc += f"\nPRIORITY {self.__priority}"

        return desc



"""
This module provides REST API handlers for scenario events.
"""
import warnings
import falcon

from .handlers import ScenarioBaseHandler
from ...simulation import Leakage, SensorFault, SensorReadingAttack
from ...simulation.events import SpeciesInjectionEvent


class ScenarioLeakageHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning leakages.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets all leakages of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_leakages = self.scenario_mgr.get(scenario_id).leakages
            self.send_json_response(resp, my_leakages)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Adds a new leakage to a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            leakage = self.load_json_data_from_request(req)
            if not isinstance(leakage, Leakage):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).add_leakage(leakage)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioQualityEventHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning quality events.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets all quality events of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_system_events = self.scenario_mgr.get(scenario_id).system_events
            my_quality_events = list(filter(lambda event: isinstance(event, SpeciesInjectionEvent),
                                            my_system_events))
            self.send_json_response(resp, my_quality_events)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Adds a new quality event to a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            quality_event = self.load_json_data_from_request(req)
            if not isinstance(quality_event, SpeciesInjectionEvent):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).add_system_event(quality_event)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioSensorReadingAttackHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning sensor reading attacks.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets all sensor reading attack events of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sensor_reading_attacks = self.scenario_mgr.get(scenario_id).sensor_reading_attacks
            self.send_json_response(resp, my_sensor_reading_attacks)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Adds a new sensor reading attack event to a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_reading_attack = self.load_json_data_from_request(req)
            if not isinstance(sensor_reading_attack, SensorReadingAttack):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).add_sensor_reading_attack(sensor_reading_attack)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR


class ScenarioSensorFaultHandler(ScenarioBaseHandler):
    """
    Class for handling GET and POST requests concerning sensor faults.
    """
    def on_get(self, _, resp: falcon.Response, scenario_id: str) -> None:
        """
        Gets all sensor faults of a given scenario.

        Parameters
        ----------
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            my_sensor_faults = self.scenario_mgr.get(scenario_id).sensor_faults
            self.send_json_response(resp, my_sensor_faults)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR

    def on_post(self, req: falcon.Request, resp: falcon.Response, scenario_id: str) -> None:
        """
        Adds a new sensor fault to a given scenario.

        Parameters
        ----------
        req : `falcon.Request <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#request>`_
            Request instance.
        resp : `falcon.Response <https://falcon.readthedocs.io/en/stable/api/request_and_response_asgi.html#response>`_
            Response instance.
        scenario_id : `str`
            UUID of the scenario.
        """
        try:
            if self.scenario_mgr.validate_uuid(scenario_id) is False:
                self.send_invalid_resource_id_error(resp)
                return

            sensor_fault = self.load_json_data_from_request(req)
            if not isinstance(sensor_fault, SensorFault):
                self.send_json_parsing_error(resp)
                return

            self.scenario_mgr.get(scenario_id).add_sensor_fault(sensor_fault)
        except Exception as ex:
            warnings.warn(str(ex))
            resp.status = falcon.HTTP_INTERNAL_SERVER_ERROR



"""
The Battle of the Leakage Detection and Isolation Methods (*BattLeDIM*) 2020, organized by
S. G. Vrachimis, D. G. Eliades, R. Taormina, Z. Kapelan, A. Ostfeld, S. Liu, M. Kyriakou,
P. Pavlou, M. Qiu, and M. M. Polycarpou, as part of the 2nd International CCWI/WDSA Joint
Conference in Beijing, China, aims at objectively comparing the performance of methods for
the detection and localization of leakage events, relying on SCADA measurements of flow and
pressure sensors installed within water distribution networks.

See https://github.com/KIOS-Research/BattLeDIM for details.

This module provides functions for loading the original BattLeDIM data set
:func:`~epyt_flow.data.benchmarks.battledim.load_data`, as well as methods for loading the scenarios
:func:`~epyt_flow.data.benchmarks.battledim.load_scenario` and pre-generated SCADA data
:func:`~epyt_flow.data.benchmarks.battledim.load_scada_data`.
The official scoring/evaluation is implemented in
:func:`~epyt_flow.data.benchmarks.battledim.compute_evaluation_score` -- i.e. those results can be
directly compared to the official leaderboard results.
Besides this, the user can choose to evaluate predictions using any other metric from
:mod:`~epyt_flow.metrics`.
"""
from typing import Any, Union
import os
import math
from datetime import datetime
import functools
import scipy
import pandas as pd
import numpy as np
from scipy.sparse import bsr_array

from .battledim_data import START_TIME_TEST, START_TIME_TRAIN, LEAKS_CONFIG_TEST, \
    LEAKS_CONFIG_TRAIN
from ..networks import load_ltown
from ...simulation.events import AbruptLeakage, IncipientLeakage, Leakage
from ...simulation import ScenarioConfig, EpanetConstants
from ...topology import NetworkTopology
from ...simulation.scada import ScadaData
from ...utils import get_temp_folder, to_seconds, create_path_if_not_exist, download_if_necessary


def __parse_leak_config(start_time: str, leaks_config: str) -> list[Leakage]:
    leakages = []
    for leak in leaks_config.splitlines():
        # Parse entry
        items = [i.strip() for i in leak.split(",")]
        leaky_pipe_id = items[0]
        leak_start_time = int((datetime.strptime(items[1], "%Y-%m-%d %H:%M") - start_time)
                              .total_seconds())
        leak_end_time = int((datetime.strptime(items[2], "%Y-%m-%d %H:%M") - start_time)
                            .total_seconds())
        leak_diameter = float(items[3])
        leak_type = items[4]
        leak_peak_time = int((datetime.strptime(items[5], "%Y-%m-%d %H:%M") - start_time)
                             .total_seconds())

        # Create leak config
        if leak_type == "incipient":
            leak = IncipientLeakage(link_id=leaky_pipe_id, diameter=leak_diameter,
                                    start_time=leak_start_time, end_time=leak_end_time,
                                    peak_time=leak_peak_time)
        elif leak_type == "abrupt":
            leak = AbruptLeakage(link_id=leaky_pipe_id, diameter=leak_diameter,
                                 start_time=leak_start_time, end_time=leak_end_time)

        leakages.append(leak)

    return leakages


def __create_labels(n_time_steps: int, return_test_scenario: bool,
                    links: list[str]) -> tuple[np.ndarray, scipy.sparse.bsr_array]:
    y = np.zeros(n_time_steps)

    start_time = START_TIME_TEST if return_test_scenario is True else START_TIME_TRAIN
    leaks_config = LEAKS_CONFIG_TEST if return_test_scenario is True else LEAKS_CONFIG_TRAIN
    leakages = __parse_leak_config(start_time, leaks_config)

    def leak_time_to_idx(t: int, round_up: bool = False):
        if round_up is False:
            return math.floor(t / 300)
        else:
            return math.ceil(t / 300)

    leak_locations_row = []
    leak_locations_col = []
    for leak in leakages:
        t_idx_start = leak_time_to_idx(leak.start_time)
        t_idx_end = leak_time_to_idx(leak.end_time, round_up=True)
        y[t_idx_start:t_idx_end] = 1

        leak_link_idx = links.index(leak.link_id)
        for t in range(t_idx_end - t_idx_start):
            leak_locations_row.append(t_idx_start + t)
            leak_locations_col.append(leak_link_idx)

    y_leak_locations = bsr_array(
        (np.ones(len(leak_locations_row)), (leak_locations_row, leak_locations_col)),
        shape=(n_time_steps, len(links)))

    return y, y_leak_locations


def compute_evaluation_score(y_leak_locations_pred: list[tuple[str, int]],
                             test_scenario: bool, verbose: bool = True) -> dict:
    """
    Evaluates the predictions (i.e. start time and location of leakages) as it was done in the
    BattLeDIM competition -- i.e. the output of this functions can be directly compared
    to the official leaderboard results.

    Parameters
    ----------
    y_leak_locations_pred : `list[tuple[str, int]]`
        Predictions of location (link/pipe ID) and start time
        (in seconds since simulation start) of leakages.
    test_scenario : `bool`
        True if the given predictions are made for the test scenario, False otherwise.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    `dict`
        Dictionary containing the true positive rate, true positives, false positives,
        false negatives, and total monetary (Euro) savings (only available if `test_scenario`
        is True).
    """
    # Original MATLAB implementation:
    # https://github.com/KIOS-Research/BattLeDIM/blob/master/Scoring%20Algorithm/Scoring_Algorithm.m

    # Scoring parameters
    dist_max = 300      # Max pipe distance for leakage detection (meters)
    cost_water = .8     # Cost of water per m3 (Euro)
    cost_crew = 500     # Max repair crew cost per assignment (Euro)

    hydraulic_time_step = to_seconds(minutes=5)

    # Get WDN topology and find minimum topological distance (using the pipe lengths)
    # between all nodes
    f_topology_in = os.path.join(get_temp_folder(), "BattLeDIM", "ltown.epytflow_topology")
    url_topology = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/BattLeDIM/" +\
        "ltown.epytflow_topology"

    download_if_necessary(f_topology_in, url_topology, verbose)
    topology = NetworkTopology.load_from_file(f_topology_in)

    all_pairs_shortest_path_length = topology.get_all_pairs_shortest_path_length()

    # Load ground truth
    sim_start_time = START_TIME_TEST if test_scenario is True else START_TIME_TRAIN
    leaks_config = LEAKS_CONFIG_TEST if test_scenario is True else LEAKS_CONFIG_TRAIN
    leakages = __parse_leak_config(sim_start_time, leaks_config)
    n_leakages = len(leakages)

    leak_demands = {}
    if test_scenario is True:
        # Download leak demands
        for leak in leakages:
            f_in = f"Leak_{leak.link_id}.xlsx"
            url = "https://raw.githubusercontent.com/KIOS-Research/BattLeDIM/master/" + \
                f"Scoring%20Algorithm/competition_leakages/{f_in}"

            f_local_in = os.path.join(get_temp_folder(), "BattLeDIM", f_in)
            download_if_necessary(f_local_in, url, verbose)

            df_leak_demand = pd.read_excel(f_local_in, sheet_name="Demand (m3_h)")
            leak_demand = df_leak_demand[leak.link_id].to_numpy()
            leak_demands[leak.link_id] = leak_demand

    # Evaluate given predictions/alarms
    total_savings = 0
    true_positives = 0
    false_positives = 0
    detected_leaks = []

    leak_data = []
    for leak in leakages:
        leak_data.append((leak.link_id, leak.start_time, leak.end_time))

    def __find_closest_leaky_pipe(link_id) -> tuple[str, float, int]:
        closest_leaky_pipe_id = None
        closest_dist = float("inf")
        closest_start_time = None
        closest_end_time = None

        node_a, node_b = topology.get_link_info(link_id)["nodes"]

        for leak_pipe_id, start_time_leak, end_time_leak in leak_data:
            link_info = topology.get_link_info(leak_pipe_id)
            end_node_a, end_node_b = link_info["nodes"]
            link_length = link_info["length"]

            dists = []
            dists.append(all_pairs_shortest_path_length[node_a][end_node_a] + .5 * link_length)
            dists.append(all_pairs_shortest_path_length[node_a][end_node_b] + .5 * link_length)
            dists.append(all_pairs_shortest_path_length[node_b][end_node_a] + .5 * link_length)
            dists.append(all_pairs_shortest_path_length[node_b][end_node_b] + .5 * link_length)
            if min(dists) < closest_dist:
                closest_dist = min(dists)
                closest_leaky_pipe_id = leak_pipe_id
                closest_start_time = start_time_leak
                closest_end_time = end_time_leak

        return closest_leaky_pipe_id, closest_dist, closest_start_time, closest_end_time

    for pipe_id, start_time in y_leak_locations_pred:
        # Check if leakages was found and if so, how far away it is from the ground truth
        leaky_pipe_dist = None
        leaky_pipe = None
        if any(pipe_id == leaky_pipe_id and start_time >= start_time_leak and
               start_time <= end_time_leak and
               pipe_id not in detected_leaks
               for leaky_pipe_id, start_time_leak, end_time_leak in leak_data):
            leaky_pipe_dist = 0
            leaky_pipe = pipe_id
        else:
            closest_leaky_pipe_id, dist, start_time_leak, end_time_leak = \
                __find_closest_leaky_pipe(pipe_id)
            if start_time >= start_time_leak and start_time <= end_time_leak:
                leaky_pipe_dist = dist
                leaky_pipe = closest_leaky_pipe_id

        # Compute score of current alarm
        if leaky_pipe is not None:
            detected_leaks.append(leaky_pipe)
            true_positives += 1

            water_saved = 0
            if leaky_pipe in leak_demands:
                leak_demand = leak_demands[leaky_pipe]
                start_time_idx = math.ceil(start_time / hydraulic_time_step)
                water_saved = np.sum(leak_demand[start_time_idx:])
            total_savings += water_saved * cost_water - (leaky_pipe_dist / dist_max) * cost_crew
        else:
            false_positives += 1
            total_savings += -1. * cost_crew

    # Compute final scores
    false_negatives = n_leakages - true_positives
    true_positive_rate = true_positives / (true_positives + false_negatives)

    return {"true_positive_rate": true_positive_rate, "true_positives": true_positives,
            "false_positives": false_positives, "false_negatives": false_negatives,
            "total_savings": total_savings if test_scenario is True else None}


def load_data(return_test_scenario: bool, download_dir: str = None, return_X_y: bool = False,
              return_features_desc: bool = False, return_leak_locations: bool = False,
              verbose: bool = True) -> Union[pd.DataFrame, Any]:
    """
    Loads the original BattLeDIM benchmark data set.
    Note that the data set exists in two different version --
    a training version and an evaluation/test version.

    Parameters
    ----------
    return_test_scenario : `bool`
        If True, the evaluation/test data set is returned, otherwise the historical
        (i.e. training) data set is returned.
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    return_X_y : `bool`, optional
        If True, the data is returned together with the labels (presence of a leakage) as
        two Numpy arrays, otherwise, the data is returned as a
        :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` instance.

        The default is False.
    return_features_desc : `bool`, optional
        If True and if `return_X_y` is True, the returned dictionary contains the
        features' descriptions (i.e. names) under the key "features_desc".

        The default is False.
    return_leak_locations : `bool`
        If True, the leak locations are returned as well --
        as an instance of `scipy.sparse.bsr_array`.

        The default is False.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    Either a `pandas.DataFrame <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_ instance or a tuple of `Numpy arrays <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        Benchmark data set.
    """
    # Download data files if necessary
    if return_test_scenario is True:
        url_data = "https://zenodo.org/records/4017659/files/2018_SCADA.xlsx?download=1"
        f_in = "2018_SCADA.xlsx"
    else:
        url_data = "https://zenodo.org/records/4017659/files/2019_SCADA.xlsx?download=1"
        f_in = "2019_SCADA.xlsx"

    download_dir = download_dir if download_dir is not None else get_temp_folder()
    download_dir = os.path.join(download_dir, "BattLeDIM")
    create_path_if_not_exist(download_dir)
    f_in = os.path.join(download_dir, f_in)

    download_if_necessary(f_in, url_data, verbose)

    # Load and parse data files
    df_pressures = pd.read_excel(f_in, sheet_name="Pressures (m)")
    df_pressures.columns = ["Timestamp"] + [f"Pressure_{n_id}" for n_id in df_pressures.columns[1:]]

    df_demands = pd.read_excel(f_in, sheet_name="Demands (L_h)")
    df_demands.columns = ["Timestamp"] + [f"Demand_{n_id}" for n_id in df_demands.columns[1:]]

    df_flows = pd.read_excel(f_in, sheet_name="Flows (m3_h)")
    df_flows.columns = ["Timestamp"] + [f"Flow_{l_id}" for l_id in df_flows.columns[1:]]

    df_levels = pd.read_excel(f_in, sheet_name="Levels (m)")
    df_levels.columns = ["Timestamp"] + [f"Level_{t_id}" for t_id in df_levels.columns[1:]]

    df_final = functools.reduce(lambda left, right: pd.merge(left, right, on="Timestamp"),
                                [df_pressures, df_flows, df_levels, df_demands])

    # Prepare and return final data
    if return_X_y is True:
        features_desc = list(df_final.columns)
        features_desc.remove("Timestamp")

        network_config = load_ltown(download_dir)
        links = network_config.sensor_config.links

        X = df_final[features_desc].to_numpy()
        y, y_leak_locations = __create_labels(X.shape[0], return_test_scenario, links)

        if return_features_desc is True:
            if return_leak_locations is True:
                return X, y, features_desc, y_leak_locations
            else:
                return X, y, features_desc
        else:
            if return_leak_locations is True:
                return X, y, y_leak_locations
            else:
                return X, y
    else:
        return df_final


def load_scada_data(return_test_scenario: bool, download_dir: str = None,
                    return_X_y: bool = False, return_leak_locations: bool = False,
                    verbose: bool = True) -> list[Union[ScadaData, Any]]:
    """
    Loads the SCADA data of the simulated BattLeDIM benchmark scenario -- note that due to
    randomness, these differ from the original data set which can be loaded by calling
    :func:`~epyt_flow.data.benchmarks.battledim.load_data`.

    .. warning::

        A large file (approx. 4GB) will be downloaded and loaded into memory --
        this might take some time.

    Parameters
    ----------
    return_test_scenario : `bool`
        If True, the evaluation/test scenario is returned, otherwise the historical
        (i.e. training) scenario is returned.
    download_dir : `str`, optional
        Path to the data files -- if None, the temp folder will be used.
        If the path does not exist, the data files will be downloaded to the given path.

        The default is None.
    return_X_y : `bool`, optional
        If True, the data is returned together with the labels (presence of a leakage) as
        two Numpy arrays, otherwise, the data is returned as a
        :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` instance.

        The default is False.
    return_leak_locations : `bool`
        If True, the leak locations are returned as well --
        as an instance of `scipy.sparse.bsr_array`.

        The default is False.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or `list[tuple[numpy.ndarray, numpy.ndarray]]`
        The simulated benchmark scenario as either a
        :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` instance or as a tuple of
        (X, y) `Numpy arrays <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        If 'return_leak_locations' is True, the leak locations are included
        as an instance of `scipy.sparse.bsr_array <https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.bsr_array.html>`_ as well.
    """
    download_dir = download_dir if download_dir is not None else get_temp_folder()

    url_data = "https://filedn.com/lumBFq2P9S74PNoLPWtzxG4/EPyT-Flow/BattLeDIM/"

    f_in = f"{'battledim_test' if return_test_scenario else 'battledim_train'}.epytflow_scada_data"
    download_if_necessary(os.path.join(download_dir, f_in), url_data + f_in, verbose)

    data = ScadaData.load_from_file(os.path.join(download_dir, f_in))

    X = data.get_data()
    y, y_leak_locations = __create_labels(X.shape[0], return_test_scenario,
                                          data.sensor_config.links)

    if return_X_y is True:
        if return_leak_locations is True:
            return X, y, y_leak_locations
        else:
            return X, y
    else:
        if return_leak_locations is True:
            return data, y_leak_locations
        else:
            return data


def load_scenario(return_test_scenario: bool, download_dir: str = None,
                  verbose: bool = True) -> ScenarioConfig:
    """
    Creates and returns the BattLeDIM scenario -- it can be either modified or
    passed directly to the simulator
    :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.

    .. note::

        Note that due to randomness, the simulation results differ from the original data set which
        can be loaded by calling :func:`~epyt_flow.data.benchmarks.battledim.load_data`.

    Parameters
    ----------
    return_test_scenario : `bool`
        If True, the evaluation/test scenario is returned, otherwise the historical
        (i.e. training) scenario is returned.
    download_dir : `str`, optional
        Path to the L-TOWN.inp file -- if None, the temp folder will be used.
        If the path does not exist, the .inp will be downloaded to the given path.

        The default is None.
    verbose : `bool`, optional
        If True, a progress bar is shown while downloading files.

        The default is True.

    Returns
    -------
    :class:`~epyt_flow.simulation.scenario_config.ScenarioConfig`
        Complete scenario configuration of the BattLeDIM benchmark scenario.
    """

    # Load L-Town network including the sensor placement
    if download_dir is not None:
        ltown_config = load_ltown(download_dir=download_dir, use_realistic_demands=True,
                                  include_default_sensor_placement=True, verbose=verbose)
    else:
        ltown_config = load_ltown(use_realistic_demands=True, include_default_sensor_placement=True,
                                  verbose=verbose)

    # Set simulation duration and other general parameters such as the demand model
    general_params = {"simulation_duration": to_seconds(days=365),    # One year
                      "hydraulic_time_step": to_seconds(minutes=5),   # 5min time steps
                      "reporting_time_step": to_seconds(minutes=5),
                      "demand_model": {"type": EpanetConstants.EN_PDA, "pressure_min": 0,
                                       "pressure_required": 0.1,
                                       "pressure_exponent": 0.5}
                      } | ltown_config.general_params

    # Add events
    start_time = START_TIME_TEST if return_test_scenario is True else START_TIME_TRAIN
    leaks_config = LEAKS_CONFIG_TEST if return_test_scenario is True else LEAKS_CONFIG_TRAIN
    leakages = __parse_leak_config(start_time, leaks_config)

    # Build final scenario
    return ScenarioConfig(f_inp_in=ltown_config.f_inp_in, general_params=general_params,
                          sensor_config=ltown_config.sensor_config, system_events=leakages)
